import importlib
import sys
import typing as t

__all__ = (
    "BookmarkCreateDocument",
    "BookmarkUpdateDocument",
    "TransferClientV2",
    "TunnelCreateDocument",
    "TunnelUpdateDocument",
)


# imports from `globus_sdk.experimental.transfer_v2` are done lazily to ensure
# that we do not eagerly import `requests` when attempting to use SDK
# components which do not need it
if t.TYPE_CHECKING:
    from .transfer_v2 import (
        BookmarkCreateDocument,
        BookmarkUpdateDocument,
        TransferClientV2,
        TunnelCreateDocument,
        TunnelUpdateDocument,
    )
else:
    _LAZY_IMPORT_TABLE = {
        "transfer_v2": {
            "BookmarkCreateDocument",
            "BookmarkUpdateDocument",
            "TransferClientV2",
            "TunnelCreateDocument",
            "TunnelUpdateDocument",
        }
    }

    def __getattr__(name: str) -> t.Any:
        for modname, items in _LAZY_IMPORT_TABLE.items():
            if name in items:
                mod = importlib.import_module("." + modname, __name__)
                value = getattr(mod, name)
                setattr(sys.modules[__name__], name, value)
                return value

        raise AttributeError(f"module {__name__} has no attribute {name}")



import pytest

import globus_sdk
from globus_sdk.experimental.gcs_collection_client import GCSCollectionClient


@pytest.mark.parametrize("attrname", ["https", "data_access", "resource_server"])
def test_class_level_scopes_access_raises_useful_attribute_error(attrname):
    with pytest.raises(
        AttributeError,
        match=(
            f"It is not valid to attempt to access the '{attrname}' attribute of "
            "the GCSCollectionClient class"
        ),
    ):
        getattr(GCSCollectionClient.scopes, attrname)


def test_instance_level_scopes_access_ok():
    client = GCSCollectionClient("foo_id", "https://example.com/foo")

    assert client.resource_server == "foo_id"
    assert client.scopes.https == globus_sdk.Scope(
        "https://auth.globus.org/scopes/foo_id/https"
    )
    assert client.scopes.data_access == globus_sdk.Scope(
        "https://auth.globus.org/scopes/foo_id/data_access"
    )


def test_default_scope_is_https():
    client = GCSCollectionClient("foo_id", "https://example.com/foo")

    assert client.default_scope_requirements == [
        globus_sdk.Scope("https://auth.globus.org/scopes/foo_id/https")
    ]


# this behavior imitates the `GCSClient`, which accepts a bare hostname and prepends the
# scheme if it is missing
def test_https_scheme_is_added_to_bare_address():
    client = GCSCollectionClient("foo_id", "example.com")

    assert client.base_url == "https://example.com"



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_ID = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="search",
        method="DELETE",
        path=f"/v1/index/{INDEX_ID}",
        json={
            "index_id": INDEX_ID,
            "acknowledged": True,
        },
        metadata={"index_id": INDEX_ID},
    ),
    delete_pending=RegisteredResponse(
        service="search",
        method="DELETE",
        path=f"/v1/index/{INDEX_ID}",
        status=409,
        json={
            "@datatype": "GError",
            "request_id": "3430ce9a5f9d929ef7682e4c58363dee",
            "status": 409,
            "@version": "2017-09-01",
            "message": (
                "Index status (delete_pending) did not match required status "
                "for this operation: open"
            ),
            "code": "Conflict.IncompatibleIndexStatus",
        },
        metadata={"index_id": INDEX_ID},
    ),
)



from __future__ import annotations

import typing as t

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

raw_data: list[dict[str, t.Any]] = [
    {
        "name": "Claptrap's Rough Riders",
        "parent_id": None,
        "id": "d3974728-6458-11e4-b72d-123139141556",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "d3974728-6458-11e4-b72d-123139141556",
                "identity_id": "ae332d86-d274-11e5-b885-b31714a110e9",
                "username": "sirosen@globusid.org",
                "role": "member",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "private",
            "group_members_visibility": "managers",
        },
    },
    {
        "name": "duke",
        "parent_id": "fdb38a24-03c1-11e3-86f7-12313809f035",
        "id": "7c580b9a-4592-11e3-a2a0-12313d2d6e7f",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "7c580b9a-4592-11e3-a2a0-12313d2d6e7f",
                "identity_id": "ae332d86-d274-11e5-b885-b31714a110e9",
                "username": "sirosen@globusid.org",
                "role": "member",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "authenticated",
            "group_members_visibility": "members",
        },
    },
    {
        "name": "kbase_users",
        "parent_id": None,
        "id": "99d2a548-7218-11e2-adc0-12313d2d6e7f",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "99d2a548-7218-11e2-adc0-12313d2d6e7f",
                "identity_id": "ae332d86-d274-11e5-b885-b31714a110e9",
                "username": "sirosen@globusid.org",
                "role": "member",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "authenticated",
            "group_members_visibility": "members",
        },
    },
    {
        "name": "connect",
        "parent_id": None,
        "id": "fdb38a24-03c1-11e3-86f7-12313809f035",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "fdb38a24-03c1-11e3-86f7-12313809f035",
                "identity_id": "ae332d86-d274-11e5-b885-b31714a110e9",
                "username": "sirosen@globusid.org",
                "role": "member",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "authenticated",
            "group_members_visibility": "members",
        },
    },
    {
        "name": "sirosen's Email Testing Group",
        "parent_id": None,
        "id": "b0d168b0-6398-11e4-ac82-12313b077182",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "b0d168b0-6398-11e4-ac82-12313b077182",
                "identity_id": "ae332d86-d274-11e5-b885-b31714a110e9",
                "username": "sirosen@globusid.org",
                "role": "admin",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "private",
            "group_members_visibility": "managers",
        },
    },
    {
        "name": "osg",
        "parent_id": "fdb38a24-03c1-11e3-86f7-12313809f035",
        "id": "80321e42-41a3-11e3-bef1-12313d2d6e7f",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "80321e42-41a3-11e3-bef1-12313d2d6e7f",
                "identity_id": "ae332d86-d274-11e5-b885-b31714a110e9",
                "username": "sirosen@globusid.org",
                "role": "member",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "authenticated",
            "group_members_visibility": "members",
        },
    },
    {
        "name": "Search Examples: Cookery",
        "parent_id": None,
        "id": "0a4dea26-44cd-11e8-847f-0e6e723ad808",
        "group_type": "regular",
        "enforce_session": False,
        "session_limit": 28800,
        "session_timeouts": {},
        "my_memberships": [
            {
                "group_id": "0a4dea26-44cd-11e8-847f-0e6e723ad808",
                "identity_id": "ae341a98-d274-11e5-b888-dbae3a8ba545",
                "username": "sirosen@globus.org@accounts.google.com",
                "role": "admin",
                "status": "active",
            }
        ],
        "policies": {
            "group_visibility": "authenticated",
            "group_members_visibility": "managers",
        },
    },
]
group_ids = [x["id"] for x in raw_data]
group_names = [x["name"] for x in raw_data]
member_ids = {
    group["id"]: [m["identity_id"] for m in group["my_memberships"]]
    for group in raw_data
}

RESPONSES = ResponseSet(
    metadata={
        "group_ids": group_ids,
        "group_names": group_names,
        "member_ids": member_ids,
    },
    default=RegisteredResponse(
        service="groups",
        path="/v2/groups/my_groups",
        json=raw_data,
    ),
)



"""
Data helper classes for constructing Transfer v2 API documents. All classes should
be Payload types, so they can be passed seamlessly to
:class:`TransferClient <globus_sdk.TransferClient>` methods without conversion.
"""

from .bookmark_documents import BookmarkCreateDocument, BookmarkUpdateDocument
from .tunnel_documents import TunnelCreateDocument, TunnelUpdateDocument

__all__ = (
    "BookmarkCreateDocument",
    "BookmarkUpdateDocument",
    "TunnelCreateDocument",
    "TunnelUpdateDocument",
)



from __future__ import annotations

import typing as t

from globus_sdk import exc
from globus_sdk._internal.guards import validators
from globus_sdk._internal.serializable import Serializable

from ._auth_requirements_error import GARE, GlobusAuthorizationParameters

V = t.TypeVar("V", bound="LegacyAuthRequirementsErrorVariant")


class LegacyAuthRequirementsErrorVariant(t.Protocol):
    """
    Protocol for errors which can be converted to a Globus Auth Requirements Error.
    """

    @classmethod
    def from_dict(cls: type[V], data: dict[str, t.Any]) -> V:
        pass

    def to_auth_requirements_error(self) -> GARE: ...


class LegacyDependentConsentRequiredAuthError(Serializable):
    """
    The dependent_consent_required error format emitted by the Globus Auth service.
    """

    def __init__(
        self,
        *,
        error: t.Literal["dependent_consent_required"],
        errors: list[dict[str, str]],
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        self.error = _validate_dependent_consent_required_literal("error", error)

        try:
            first_error = errors[0]
            if not isinstance(first_error, dict):
                raise TypeError
        except (AttributeError, KeyError, TypeError) as _error:
            msg = "'errors' must be list of errors with at least one error object"
            raise exc.ValidationError(msg) from _error
        self.errors = errors
        self.unapproved_scopes = validators.str_list(
            "unapproved_scopes", first_error.get("unapproved_scopes", [])
        )
        self.extra = extra or {}

    def to_auth_requirements_error(self) -> GARE:
        """
        Return a GlobusAuthRequirementsError representing this error.
        """

        return GARE(
            code="ConsentRequired",
            authorization_parameters=GlobusAuthorizationParameters(
                required_scopes=self.unapproved_scopes,
            ),
        )


class LegacyConsentRequiredTransferError(Serializable):
    """
    The ConsentRequired error format emitted by the Globus Transfer service.
    """

    def __init__(
        self,
        *,
        code: t.Literal["ConsentRequired"],
        required_scopes: list[str],
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        self.code = _validate_consent_required_literal("code", code)
        self.required_scopes = validators.str_list("required_scopes", required_scopes)
        self.extra = extra or {}

    def to_auth_requirements_error(self) -> GARE:
        """
        Return a GlobusAuthRequirementsError representing this error.
        """
        return GARE(
            code=self.code,
            authorization_parameters=GlobusAuthorizationParameters(
                required_scopes=self.required_scopes,
                session_message=self.extra.get("message"),
            ),
            extra=self.extra,
        )


class LegacyConsentRequiredAPError(Serializable):
    """
    The ConsentRequired error format emitted by the legacy Globus Transfer
    Action Providers.
    """

    def __init__(
        self,
        *,
        code: t.Literal["ConsentRequired"],
        required_scope: str,
        extra: dict[str, t.Any] | None,
    ) -> None:
        self.code = _validate_consent_required_literal("code", code)
        self.required_scope = validators.str_("required_scope", required_scope)
        self.extra = extra or {}

    def to_auth_requirements_error(self) -> GARE:
        """
        Return a GlobusAuthRequirementsError representing this error.

        Normalizes the required_scope field to a list and uses the description
        to set the session message.
        """
        return GARE(
            code=self.code,
            authorization_parameters=GlobusAuthorizationParameters(
                required_scopes=[self.required_scope],
                session_message=self.extra.get("description"),
                extra=self.extra.get("authorization_parameters"),
            ),
            extra={
                k: v for k, v in self.extra.items() if k != "authorization_parameters"
            },
        )


class LegacyAuthorizationParameters(Serializable):
    """
    An Authorization Parameters object that describes all known variants in use by
    Globus services.
    """

    def __init__(
        self,
        *,
        session_message: str | None = None,
        session_required_identities: list[str] | None = None,
        session_required_policies: str | list[str] | None = None,
        session_required_single_domain: str | list[str] | None = None,
        session_required_mfa: bool | None = None,
        prompt: t.Literal["login"] | None = None,
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        self.session_message = validators.opt_str("session_message", session_message)
        self.session_required_identities = validators.opt_str_list(
            "session_required_identities", session_required_identities
        )
        self.session_required_policies = validators.opt_str_list_or_commasep(
            "session_required_policies", session_required_policies
        )
        self.session_required_single_domain = validators.opt_str_list_or_commasep(
            "session_required_single_domain", session_required_single_domain
        )
        self.session_required_mfa = validators.opt_bool(
            "session_required_mfa", session_required_mfa
        )
        if prompt in [None, "login"]:
            self.prompt = prompt
        else:
            raise exc.ValidationError("'prompt' must be 'login' or null")
        self.extra = extra or {}

    def to_authorization_parameters(
        self,
    ) -> GlobusAuthorizationParameters:
        """
        Return a normalized GlobusAuthorizationParameters instance representing
        these parameters.

        Normalizes fields that may have been provided
        as comma-delimited strings to lists of strings.
        """
        return GlobusAuthorizationParameters(
            session_message=self.session_message,
            session_required_identities=self.session_required_identities,
            session_required_mfa=self.session_required_mfa,
            session_required_policies=self.session_required_policies,
            session_required_single_domain=self.session_required_single_domain,
            prompt=self.prompt,
            extra=self.extra,
        )


class LegacyAuthorizationParametersError(Serializable):
    """
    Defines an Authorization Parameters error that describes all known variants
    in use by Globus services.
    """

    DEFAULT_CODE = "AuthorizationRequired"

    def __init__(
        self,
        *,
        authorization_parameters: dict[str, t.Any] | LegacyAuthorizationParameters,
        code: str | None = None,
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        # Apply default, if necessary
        self.code = validators.str_("code", code or self.DEFAULT_CODE)
        self.authorization_parameters = validators.instance_or_dict(
            "authorization_parameters",
            authorization_parameters,
            LegacyAuthorizationParameters,
        )
        self.extra = extra or {}

    def to_auth_requirements_error(self) -> GARE:
        """
        Return a GlobusAuthRequirementsError representing this error.
        """
        authorization_parameters = (
            self.authorization_parameters.to_authorization_parameters()
        )
        return GARE(
            authorization_parameters=authorization_parameters,
            code=self.code,
            extra=self.extra,
        )


def _validate_consent_required_literal(
    name: str, value: t.Any
) -> t.Literal["ConsentRequired"]:
    if value == "ConsentRequired":
        return "ConsentRequired"
    raise exc.ValidationError(f"'{name}' must be the string 'ConsentRequired'")


def _validate_dependent_consent_required_literal(
    name: str, value: t.Any
) -> t.Literal["dependent_consent_required"]:
    if value == "dependent_consent_required":
        return "dependent_consent_required"
    msg = f"'{name}' must be the string 'dependent_consent_required'"
    raise exc.ValidationError(msg)



import globus_sdk
from globus_sdk.testing import load_response


def test_submit(compute_client_v3: globus_sdk.ComputeClientV3):
    meta = load_response(compute_client_v3.submit).metadata
    submit_doc = {
        "tasks": {
            meta["function_id"]: [
                "36\n00\ngASVDAAAAAAAAACMBlJvZG5leZSFlC4=\n12 ...",
                "36\n00\ngASVCwAAAAAAAACMBUJvYmJ5lIWULg==\n12 ...",
            ],
        },
        "task_group_id": meta["task_group_id"],
        "create_queue": True,
        "user_runtime": {
            "globus_compute_sdk_version": "2.29.0",
            "globus_sdk_version": "3.46.0",
            "python_version": "3.11.9",
        },
    }

    res = compute_client_v3.submit(endpoint_id=meta["endpoint_id"], data=submit_doc)

    assert res.http_status == 200
    assert res.data["request_id"] == meta["request_id"]
    assert res.data["task_group_id"] == meta["task_group_id"]
    assert res.data["endpoint_id"] == meta["endpoint_id"]
    assert res.data["tasks"] == {
        meta["function_id"]: [meta["task_id"], meta["task_id_2"]]
    }



from __future__ import annotations

import abc
import time
import typing as t

import globus_sdk
from globus_sdk.authorizers import (
    AccessTokenAuthorizer,
    ClientCredentialsAuthorizer,
    GlobusAuthorizer,
    RefreshTokenAuthorizer,
)
from globus_sdk.services.auth import OAuthTokenResponse
from globus_sdk.token_storage import ValidatingTokenStorage
from globus_sdk.token_storage.validating_token_storage import MissingTokenError

GA = t.TypeVar("GA", bound=GlobusAuthorizer)


class AuthorizerFactory(
    t.Generic[GA],
    metaclass=abc.ABCMeta,
):
    """
    An ``AuthorizerFactory`` is an interface for getting some class of
    ``GlobusAuthorizer`` from a ``ValidatingTokenStorage`` that meets the
    authorization requirements used to initialize the ``ValidatingTokenStorage``.

    An ``AuthorizerFactory`` keeps a cache of authorizer objects that are
    reused until its ``store_token_response`` method is called.
    """

    def __init__(self, token_storage: ValidatingTokenStorage) -> None:
        """
        :param token_storage: The ``ValidatingTokenStorage`` used
        for defining and validating the set of authorization requirements that
        constructed authorizers will meet and accessing underlying token storage
        """
        self.token_storage = token_storage
        self._authorizer_cache: dict[str, GA] = {}

    def store_token_response_and_clear_cache(
        self, token_res: OAuthTokenResponse
    ) -> None:
        """
        Store a token response in the underlying ``ValidatingTokenStorage``
        and clear the authorizer cache.

        This should not be called when a ``RenewingAuthorizer`` created by this factory
        gets new tokens for itself as there is no need to clear the cache.

        :param token_res: An ``OAuthTokenResponse`` containing token data to be stored
            in the underlying ``ValidatingTokenStorage``.
        """
        self.token_storage.store_token_response(token_res)
        self.clear_cache()

    def clear_cache(self, *resource_servers: str) -> None:
        """
        Clear the authorizer cache for the given resource servers.
        If no resource servers are given, clear the entire cache.

        :param resource_servers: The resource servers for which to clear the cache
        """
        if not resource_servers:
            self._authorizer_cache = {}
        else:
            for resource_server in resource_servers:
                if resource_server in self._authorizer_cache:
                    del self._authorizer_cache[resource_server]

    def get_authorizer(self, resource_server: str) -> GA:
        """
        Either retrieve a cached authorizer for the given resource server or construct
        a new one if none is cached.

        :param resource_server: The resource server the authorizer will produce
            authentication for

        :raises: :exc:`MissingTokenError` if the underlying ``TokenStorage`` does not
            have any token data for the given resource server.
        :raises: :exc:`UnmetScopeRequirementsError` if the stored token data does not
            meet the scope requirements for the given resource server.
        :returns: A ``GlobusAuthorizer`` for the given resource server
        """
        if resource_server in self._authorizer_cache:
            return self._authorizer_cache[resource_server]

        new_authorizer = self._make_authorizer(resource_server)
        self._authorizer_cache[resource_server] = new_authorizer
        return new_authorizer

    @abc.abstractmethod
    def _make_authorizer(self, resource_server: str) -> GA:
        """
        Construct the ``GlobusAuthorizer`` class specific to this ``AuthorizerFactory``

        :param resource_server: The resource server the authorizer will produce
            authentication for
        """


class AccessTokenAuthorizerFactory(AuthorizerFactory[AccessTokenAuthorizer]):
    """
    An ``AuthorizerFactory`` that constructs ``AccessTokenAuthorizer``.
    """

    def __init__(self, token_storage: ValidatingTokenStorage) -> None:
        super().__init__(token_storage)
        self._cached_authorizer_expiration: dict[str, int] = {}

    def store_token_response_and_clear_cache(
        self, token_res: OAuthTokenResponse
    ) -> None:
        super().store_token_response_and_clear_cache(token_res)
        self._cached_authorizer_expiration = {}

    def clear_cache(self, *resource_servers: str) -> None:
        if not resource_servers:
            self._cached_authorizer_expiration = {}
        else:
            for resource_server in resource_servers:
                if resource_server in self._cached_authorizer_expiration:
                    del self._cached_authorizer_expiration[resource_server]

        super().clear_cache(*resource_servers)

    def get_authorizer(self, resource_server: str) -> AccessTokenAuthorizer:
        """
        Either retrieve a cached authorizer for the given resource server or construct
        a new one if none is cached.

        :param resource_server: The resource server the authorizer will produce
            authentication for

        :raises: :exc:`MissingTokenError` if the underlying ``TokenStorage`` does not
            have any token data for the given resource server.
        :raises: :exc:`UnmetScopeRequirementsError` if the stored token data does not
            meet the scope requirements for the given resource server.
        :raises: :exc:`ExpiredTokenError` if the stored access token for the given
            resource server has expired.
        :returns: An ``AccessTokenAuthorizer`` for the given resource server
        """

        if resource_server in self._cached_authorizer_expiration:
            if self._cached_authorizer_expiration[resource_server] < time.time():
                del self._cached_authorizer_expiration[resource_server]
                del self._authorizer_cache[resource_server]

        return super().get_authorizer(resource_server)

    def _make_authorizer(self, resource_server: str) -> AccessTokenAuthorizer:
        """
        Construct an ``AccessTokenAuthorizer`` for the given resource server.

        :param resource_server: The resource server the authorizer will produce
            authentication for
        :raises: :exc:`ExpiredTokenError` if the stored access token for the given
            resource server has expired
        """
        token_data = self.token_storage.get_token_data(resource_server)
        return AccessTokenAuthorizer(token_data.access_token)


class RefreshTokenAuthorizerFactory(AuthorizerFactory[RefreshTokenAuthorizer]):
    """
    An ``AuthorizerFactory`` that constructs ``RefreshTokenAuthorizer``.
    """

    def __init__(
        self,
        token_storage: ValidatingTokenStorage,
        auth_login_client: globus_sdk.AuthLoginClient,
    ) -> None:
        """
        :param token_storage: The ``ValidatingTokenStorage`` used
        for defining and validating the set of authorization requirements that
        constructed authorizers will meet and accessing underlying token storage
        :auth_login_client: The ``AuthLoginCLient` used for refreshing tokens with
            Globus Auth
        """
        super().__init__(token_storage)
        self.auth_login_client = auth_login_client

    def _make_authorizer(self, resource_server: str) -> RefreshTokenAuthorizer:
        """
        Construct a ``RefreshTokenAuthorizer`` for the given resource server.

        :param resource_server: The resource server the authorizer will produce
            authentication for
        :raises: :exc:`MissingTokenError` if the stored token data for the given
            resource server does not have a refresh token
        """
        token_data = self.token_storage.get_token_data(resource_server)
        # this condition should be unreachable -- `get_token_data` will invoke the
        # `refresh_token` validator
        # the only way to reach it would be to manipulate the validators such that
        # the check is removed
        if token_data.refresh_token is None:  # pragma: no cover
            raise ValueError("token data missing required field refresh_token")
        return RefreshTokenAuthorizer(
            refresh_token=token_data.refresh_token,
            auth_client=self.auth_login_client,
            access_token=token_data.access_token,
            expires_at=token_data.expires_at_seconds,
            on_refresh=self.token_storage.store_token_response,
        )


class ClientCredentialsAuthorizerFactory(
    AuthorizerFactory[ClientCredentialsAuthorizer]
):
    """
    An ``AuthorizerFactory`` that constructs ``ClientCredentialsAuthorizer``.

    ClientCredentialAuthorizers are a special flavor of RenewingAuthorizers which
      use the client credentials grant type and a refresh token to keep up-to-date
      access tokens for a resource server.
    """

    def __init__(
        self,
        token_storage: ValidatingTokenStorage,
        confidential_client: globus_sdk.ConfidentialAppAuthClient,
        scope_requirements: dict[str, list[globus_sdk.Scope]],
    ) -> None:
        """
        :param token_storage: The ``ValidatingTokenStorage`` used
        for defining and validating the set of authorization requirements that
        constructed authorizers will meet and accessing underlying token storage
        :param confidential_client: The ``ConfidentialAppAuthClient`` that will
            get client credentials tokens from Globus Auth to act as itself
        """
        self.confidential_client = confidential_client
        self.scope_requirements = scope_requirements
        super().__init__(token_storage)

    def _make_authorizer(
        self,
        resource_server: str,
    ) -> ClientCredentialsAuthorizer:
        """
        Construct a ``ClientCredentialsAuthorizer`` for the given resource server.

        Does not require that tokens exist in the token storage but will use them if
        present.

        :param resource_server: The resource server the authorizer will produce
            authentication for. The ``ValidatingTokenStorage`` used to create the
            ``ClientCredentialsAuthorizerFactory`` must have scope requirements defined
            for this resource server.
        """
        scopes = self.scope_requirements.get(resource_server)
        if scopes is None:
            raise ValueError(
                "ValidatingTokenStorage has no scope_requirements for "
                f"resource_server {resource_server}"
            )

        try:
            token_data = self.token_storage.get_token_data(resource_server)
            access_token = token_data.access_token
            expires_at = token_data.expires_at_seconds
        except MissingTokenError:
            access_token, expires_at = None, None

        return ClientCredentialsAuthorizer(
            confidential_client=self.confidential_client,
            scopes=scopes,
            access_token=access_token,
            expires_at=expires_at,
            on_refresh=self.token_storage.store_token_response,
        )



import textwrap


def test_addcontent_generating_text(
    sphinxext, docutils_runner, register_temporary_directive
):

    class MyDirective(sphinxext.directives.AddContentDirective):
        def gen_rst(self):
            yield "a"
            yield "b"

    register_temporary_directive("mydirective", MyDirective)

    etree = docutils_runner.to_etree(".. mydirective::")

    assert etree.tag == "document"
    assert etree.get("source") == "TEST"
    paragraph_element = etree.find("paragraph")
    assert paragraph_element is not None
    assert paragraph_element.text == textwrap.dedent("""\
        a
        b""")


def test_addcontent_generating_warning(
    sphinxext, docutils_runner, register_temporary_directive
):

    class MyDirective(sphinxext.directives.AddContentDirective):
        def gen_rst(self):
            yield ".. note::"
            yield ""
            yield "     Some note content here."
            yield "     Multiline."

    register_temporary_directive("mydirective", MyDirective)

    etree = docutils_runner.to_etree(".. mydirective::")

    assert etree.tag == "document"
    assert etree.get("source") == "TEST"
    note_element = etree.find("note")
    assert note_element is not None
    paragraph_element = note_element.find("paragraph")
    assert paragraph_element is not None
    assert paragraph_element.text == "Some note content here.\nMultiline."



import uuid

zero_id = uuid.UUID(int=0)
zero_id_s = str(zero_id)


def test_manage_collections_scope_helper(client):
    sc = client.get_gcs_endpoint_scopes(zero_id)
    assert (
        str(sc.manage_collections)
        == f"urn:globus:auth:scope:{zero_id_s}:manage_collections"
    )
    # data_access is separated from endpoint scopes
    assert not hasattr(sc, "data_access")


def test_data_access_scope_helper(client):
    sc = client.get_gcs_collection_scopes(zero_id)
    assert (
        str(sc.data_access) == f"https://auth.globus.org/scopes/{zero_id_s}/data_access"
    )
    assert str(sc.https) == f"https://auth.globus.org/scopes/{zero_id_s}/https"
    # manage_collections is separated from collection scopes
    assert not hasattr(sc, "manage_collections")


def test_contains_scope_properties(client):
    ep_sc = client.get_gcs_endpoint_scopes(zero_id)
    assert ep_sc.manage_collections in list(ep_sc)

    collection_sc = client.get_gcs_collection_scopes(zero_id)
    assert collection_sc.data_access in list(collection_sc)



import globus_sdk.gare


def coalesce(
    *gares: globus_sdk.gare.GARE,
    session_message: str | None = None,
    prompt: str | None = None,
) -> list[globus_sdk.gare.GARE]:
    # build a list of GARE fields which are allowed to merge
    safe_fields = ["session_required_policies", "required_scopes"]
    if session_message is not None:
        safe_fields.append("session_message")
    if prompt is not None:
        safe_fields.append("prompt")

    # Build lists of GAREs that can and cannot be merged
    candidates, non_candidates = [], []
    for g in gares:
        if _is_candidate(g, safe_fields):
            candidates.append(g)
        else:
            non_candidates.append(g)

    # if no GAREs were safe to merge, return early
    if not candidates:
        return non_candidates

    # merge safe GAREs and override any provided field values
    combined = _safe_combine(candidates)
    if session_message is not None:
        combined.authorization_parameters.session_message = session_message
    if prompt is not None:
        combined.authorization_parameters.prompt = prompt

    # return the reduced list of GAREs
    return [combined] + non_candidates


def _is_candidate(g: globus_sdk.gare.GARE, safe_fields: list[str]) -> bool:
    params = g.authorization_parameters

    # check all of the supported GARE fields
    for field_name in (
        "session_message",
        "session_required_identities",
        "session_required_policies",
        "session_required_single_domain",
        "session_required_mfa",
        "required_scopes",
        "prompt",
    ):
        # if the field is considered safe, ignore it
        if field_name in safe_fields:
            continue
        # if the field isn't considered safe and it is set,
        # then the GARE shouldn't be merged
        if getattr(params, field_name) is not None:
            return False

    # if we didn't find any invalidating fields, it must be safe to merge
    return True


def _safe_combine(mergeable_gares: list[globus_sdk.gare.GARE]) -> globus_sdk.gare.GARE:
    code = "AuthorizationRequired"
    if all(g.code == "ConsentRequired" for g in mergeable_gares):
        code = "ConsentRequired"

    combined_params = globus_sdk.gare.GlobusAuthorizationParameters(
        session_required_policies=_concat(
            [
                g.authorization_parameters.session_required_policies
                for g in mergeable_gares
            ]
        ),
        required_scopes=_concat(
            [g.authorization_parameters.required_scopes for g in mergeable_gares]
        ),
    )

    return globus_sdk.gare.GARE(code=code, authorization_parameters=combined_params)


def _concat(values: list[list[str] | None]) -> list[str] | None:
    if all(v is None for v in values):
        return None

    return [element for value in values if value is not None for element in value]


if __name__ == "__main__":
    # these are example errors
    case1 = globus_sdk.gare.to_gare(
        {
            "code": "ConsentRequired",
            "authorization_parameters": {"required_scopes": ["foo"]},
        }
    )
    case2 = globus_sdk.gare.to_gare(
        {
            "code": "ConsentRequired",
            "authorization_parameters": {"required_scopes": ["bar"]},
        }
    )
    case3 = globus_sdk.gare.to_gare(
        {
            "code": "AuthorizationRequired",
            "authorization_parameters": {"required_scopes": ["baz"]},
        }
    )
    case4 = globus_sdk.gare.to_gare(
        {
            "code": "AuthorizationRequired",
            "authorization_parameters": {
                "session_required_policies": [
                    "f2047039-2f07-4f13-b21b-b2edf7f9d329",
                    "2fc6d9a3-9322-48a1-ad39-5dcf63a593a7",
                ],
            },
        }
    )
    case5 = globus_sdk.gare.to_gare(
        {
            "code": "AuthorizationRequired",
            "authorization_parameters": {
                "session_required_policies": [
                    "f2047039-2f07-4f13-b21b-b2edf7f9d329",
                    "2fc6d9a3-9322-48a1-ad39-5dcf63a593a7",
                ],
                "session_required_mfa": True,
            },
        }
    )
    case6 = globus_sdk.gare.to_gare(
        {
            "code": "AuthorizationRequired",
            "authorization_parameters": {
                "session_required_policies": ["ba10f6f1-5b23-4703-bfb7-4fdd7b529546"],
                "session_message": "needs a policy",
            },
        }
    )
    case7 = globus_sdk.gare.to_gare(
        {
            "code": "AuthorizationRequired",
            "authorization_parameters": {
                "session_required_policies": ["ba10f6f1-5b23-4703-bfb7-4fdd7b529546"],
                "prompt": "login",
            },
        }
    )

    print("\n--full merge--\n")
    print("\ncombining two:")
    for g in coalesce(case1, case2):
        print(" -", g)
    print("\ncombining three:")
    for g in coalesce(case1, case2, case3):
        print(" -", g)
    print("\ncombining four:")
    for g in coalesce(case1, case2, case3, case4):
        print(" -", g)

    print("\n--no merge--\n")
    print("\ncombining two:")
    for g in coalesce(case1, case5):
        print(" -", g)
    print("\ncombining two:")
    for g in coalesce(case4, case5):
        print(" -", g)
    print("\ncombining two:")
    for g in coalesce(case2, case6):
        print(" -", g)
    print("\ncombining two:")
    for g in coalesce(case4, case7):
        print(" -", g)

    print("\n--merge due to explicit param--\n")
    print("\ncombining two:")
    for g in coalesce(case1, case6, session_message="explicit message"):
        print(" -", g)
    print("\ncombining two:")
    for g in coalesce(case4, case7, prompt="login"):
        print(" -", g)



import pytest

from globus_sdk.scopes import ScopeParser


def _make_deep_scope(depth):
    big_scope = ""
    for i in range(depth):
        big_scope += f"foo{i}["
    big_scope += "bar"
    for _ in range(depth):
        big_scope += "]"
    return big_scope


def _make_wide_scope(width):
    big_scope = ""
    for i in range(width):
        big_scope += f"foo{i} "
    return big_scope


@pytest.mark.parametrize("depth", (10, 100, 1000, 2000, 3000, 4000, 5000))
def test_deep_scope_parsing(benchmark, depth):
    scope_string = _make_deep_scope(depth)
    benchmark(ScopeParser.parse, scope_string)


@pytest.mark.parametrize("width", (5000, 10000))
def test_wide_scope_parsing(benchmark, width):
    scope_string = _make_wide_scope(width)
    benchmark(ScopeParser.parse, scope_string)



"""
WARNING: for internal use only.
Everything in SDK private modules is meant to be internal only, but that
holds for this module **in particular**.

Usage:

    from globus_sdk._internal.classprop import classproperty

    class A:
        @classproperty
        def foo(self_or_cls): ...
"""

from __future__ import annotations

import os
import sys
import typing as t

T = t.TypeVar("T")
R = t.TypeVar("R")


def _in_sphinx_build() -> bool:  # pragma: no cover
    # check if `sphinx-build` was used to invoke
    return os.path.basename(sys.argv[0]) in ["sphinx-build", "sphinx-build.exe"]


class _classproperty(t.Generic[T, R]):
    """
    A hybrid class/instance property descriptor.

    On a class, the decorated method will be invoked with `cls`.
    On an instance, the decorated method will be invoked with `self`.
    """

    def __init__(self, func: t.Callable[[type[T] | T], R]) -> None:
        self.func = func

    def __get__(self, obj: T | None, cls: type[T]) -> R:
        # NOTE: our __get__ here prefers the object over the class when possible
        # although well-defined behavior for a descriptor, this contradicts the
        # expectation that developers may have from `classmethod`
        if obj is None:
            return self.func(cls)
        return self.func(obj)


# if running under sphinx, define this as the stacked classmethod(property(...))
# decoration, so that proper autodoc generation happens
# this is based on the python3.9 behavior which supported stacking these decorators
# however, that support was pulled in 3.10 and is not going to be reintroduced at
# present
# therefore, this sphinx behavior may not be stable in the long term
if _in_sphinx_build():  # pragma: no cover

    def classproperty(func: t.Callable[[T | type[T]], R]) -> _classproperty[T, R]:
        # type ignore this because
        # - it doesn't match the return type
        # - mypy doesn't understand classmethod(property(...)) on older pythons
        return classmethod(property(func))  # type: ignore

else:

    def classproperty(func: t.Callable[[T | type[T]], R]) -> _classproperty[T, R]:
        return _classproperty(func)



import datetime
import json

import pytest

from globus_sdk import TimerJob, TimersAPIError
from globus_sdk.testing import get_last_request, load_response


def test_list_jobs(client):
    meta = load_response(client.list_jobs).metadata
    response = client.list_jobs()
    assert response.http_status == 200
    assert set(meta["job_ids"]) == {job["job_id"] for job in response.data["jobs"]}


def test_get_job(client):
    meta = load_response(client.get_job).metadata
    response = client.get_job(meta["job_id"])
    assert response.http_status == 200
    assert response.data.get("job_id") == meta["job_id"]


def test_get_job_errors(client):
    meta = load_response(client.get_job, case="simple_500_error").metadata
    with pytest.raises(TimersAPIError) as excinfo:
        client.get_job(meta["job_id"])
    err = excinfo.value
    assert err.http_status == 500
    assert err.code == "ERROR"
    assert err.message == "Request failed terribly"


@pytest.mark.parametrize("start", [datetime.datetime.now(), "2022-04-05T06:00:00"])
@pytest.mark.parametrize(
    "interval", [datetime.timedelta(days=1), datetime.timedelta(minutes=60), 600, None]
)
def test_create_job(client, start, interval):
    meta = load_response(client.create_job).metadata
    timer_job = TimerJob(
        "https://example.bogus/bogus-callback", {"bogus": "bogus_body"}, start, interval
    )

    response = client.create_job(timer_job)
    assert response.http_status == 201
    assert response.data["job_id"] == meta["job_id"]

    req_body = json.loads(get_last_request().body)
    if isinstance(start, datetime.datetime):
        assert req_body["start"] == start.isoformat()
    else:
        assert req_body["start"] == start
    if isinstance(interval, datetime.timedelta):
        assert req_body["interval"] == interval.total_seconds()
    else:
        assert req_body["interval"] == interval
    assert req_body["callback_url"] == "https://example.bogus/bogus-callback"


def test_create_job_validation_error(client):
    meta = load_response(client.create_job, case="validation_error").metadata
    timer_job = TimerJob(
        "https://example.bogus/bogus-callback",
        {"bogus": "bogus_body"},
        "2022-04-05T06:00:00",
        1800,
    )

    with pytest.raises(TimersAPIError) as excinfo:
        client.create_job(timer_job)

    err = excinfo.value
    assert err.http_status == 422
    assert err.code is None
    assert err.messages == meta["expect_messages"]


def test_update_job(client):
    meta = load_response(client.update_job).metadata
    response = client.update_job(meta["job_id"], {"name": meta["name"]})
    assert response.http_status == 200
    assert response.data["job_id"] == meta["job_id"]
    assert response.data["name"] == meta["name"]


def test_delete_job(client):
    meta = load_response(client.delete_job).metadata
    response = client.delete_job(meta["job_id"])
    assert response.http_status == 200
    assert response.data["job_id"] == meta["job_id"]


def test_pause_job(client):
    meta = load_response(client.pause_job).metadata
    response = client.pause_job(meta["job_id"])
    assert response.http_status == 200
    assert "Successfully paused" in response.data["message"]


@pytest.mark.parametrize("update_credentials", [True, False, None])
def test_resume_job(update_credentials, client):
    meta = load_response(client.resume_job).metadata

    kwargs = {}
    if update_credentials is not None:
        kwargs["update_credentials"] = update_credentials

    response = client.resume_job(meta["job_id"], **kwargs)
    assert response.http_status == 200
    assert json.loads(response._raw_response.request.body) == kwargs
    assert "Successfully resumed" in response.data["message"]



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_DOC

_two_hop_transfer_create_request = {
    k: TWO_HOP_TRANSFER_FLOW_DOC[k]
    for k in [
        "title",
        "definition",
        "input_schema",
        "subtitle",
        "description",
        "flow_viewers",
        "flow_starters",
        "flow_administrators",
        "run_managers",
        "run_monitors",
        "keywords",
    ]
}
RESPONSES = ResponseSet(
    metadata={
        "params": _two_hop_transfer_create_request,
    },
    default=RegisteredResponse(
        service="flows",
        path="/flows",
        method="POST",
        json=TWO_HOP_TRANSFER_FLOW_DOC,
    ),
    bad_admin_principal_error=RegisteredResponse(
        service="flows",
        path="/flows",
        method="POST",
        status=422,
        json={
            "error": {
                "code": "UNPROCESSABLE_ENTITY",
                "detail": [
                    {
                        "loc": ["flow_administrators", 0],
                        "msg": (
                            "Unrecognized principal string: "
                            '"ae341a98-d274-11e5-b888-dbae3a8ba545". '
                            'Allowed principal types in role "FlowAdministrator": '
                            "[<AuthGroupUrn>, <AuthIdentityUrn>]"
                        ),
                        "type": "value_error",
                    },
                    {
                        "loc": ["flow_administrators", 1],
                        "msg": (
                            "Unrecognized principal string: "
                            '"4fab4345-6d20-43a0-9a25-16b2e3bbe765". '
                            'Allowed principal types in role "FlowAdministrator": '
                            "[<AuthGroupUrn>, <AuthIdentityUrn>]"
                        ),
                        "type": "value_error",
                    },
                ],
            },
            "debug_id": "cf71b1d1-ab7e-48b1-8c54-764201d28ded",
        },
    ),
)



from __future__ import annotations

import typing as t
import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

INDEX_IDS = [str(uuid.uuid1()), str(uuid.uuid1())]
INDEX_ATTRIBUTES: dict[str, dict[str, t.Any]] = {
    INDEX_IDS[0]: {
        "subscription_id": None,
        "creation_date": "2038-07-17 16:48:24",
        "display_name": "Index of Indexed Awesomeness",
        "description": "Turbo Awesome",
        "max_size_in_mb": 1,
        "size_in_mb": 0,
        "num_subjects": 0,
        "num_entries": 0,
        "permissions": ["owner"],
    },
    INDEX_IDS[1]: {
        "subscription_id": str(uuid.uuid1()),
        "creation_date": "2470-10-11 20:09:40",
        "display_name": "Catalog of encyclopediae",
        "description": "Encyclopediae from Britannica to Wikipedia",
        "max_size_in_mb": 100,
        "size_in_mb": 23,
        "num_subjects": 1822,
        "num_entries": 3644,
        "permissions": ["writer"],
    },
}


def _make_doc(index_id: str) -> dict[str, t.Any]:
    attrs = INDEX_ATTRIBUTES[index_id]
    return {
        "@datatype": "GSearchIndex",
        "@version": "2017-09-01",
        "id": index_id,
        "is_trial": False if attrs["subscription_id"] else True,
        "status": "open",
        **attrs,
    }


RESPONSES = ResponseSet(
    metadata={"index_ids": INDEX_IDS, "index_attributes": INDEX_ATTRIBUTES},
    default=RegisteredResponse(
        service="search",
        path="/v1/index_list",
        json={
            "index_list": [
                _make_doc(INDEX_IDS[0]),
                _make_doc(INDEX_IDS[1]),
            ]
        },
    ),
)



import json

from globus_sdk.testing import get_last_request, load_response


def test_update_tunnel(client):
    meta = load_response(client.update_tunnel).metadata

    label = "New Name"
    update_doc = {
        "data": {
            "type": "Tunnel",
            "attributes": {
                "label": "New Name",
            },
        }
    }
    res = client.update_tunnel(meta["tunnel_id"], update_doc)
    assert res.http_status == 200
    assert res["data"]["type"] == "Tunnel"

    req = get_last_request()
    sent = json.loads(req.body)
    assert sent["data"]["attributes"]["label"] == label



"""
Data helper classes for constructing Transfer API documents. All classes should
be Payload types, so they can be passed seamlessly to
:class:`TransferClient <globus_sdk.TransferClient>` methods without conversion.
"""

from .delete_data import DeleteData
from .transfer_data import TransferData
from .tunnel_data import CreateTunnelData

__all__ = ("TransferData", "DeleteData", "CreateTunnelData")



"""
Custom retry check collection for the TransferClientV2 that overrides
the default check_transient_error
"""

from __future__ import annotations

from globus_sdk.transport import RetryCheck, RetryCheckResult, RetryContext
from globus_sdk.transport.default_retry_checks import (
    DEFAULT_RETRY_CHECKS,
    check_transient_error,
)


def check_transfer_v2_transient_error(ctx: RetryContext) -> RetryCheckResult:
    """
    check for transient error status codes which could be resolved by
    retrying the request. Does not retry ExternalErrors or EndpointErrors
    as those are unlikely to actually be transient.

    :param ctx: The context object which describes the state of the request and the
        retries which may already have been attempted
    """
    # JSON:API uses "the most generally applicable HTTP error code" for the
    # status of the response, so if its a transient status code then all of the
    # error objects should also have transient status codes.
    retry_config = ctx.caller_info.retry_config
    if ctx.response is not None and (
        ctx.response.status_code in retry_config.transient_error_status_codes
    ):
        try:
            # if any of the error objects have a `code` of ExternalError or
            # EndpointError then the error likely isn't transient
            errors = ctx.response.json()["errors"]
            for error in errors:
                if error["code"] in ("ExternalError", "EndpointError"):
                    return RetryCheckResult.no_decision

        except (ValueError, KeyError):
            pass

        return RetryCheckResult.do_retry

    return RetryCheckResult.no_decision


# Transfer retry checks are the defaults with the transient error one replaced
TRANSFER_V2_DEFAULT_RETRY_CHECKS: tuple[RetryCheck, ...] = tuple(
    check_transfer_v2_transient_error if check is check_transient_error else check
    for check in DEFAULT_RETRY_CHECKS
)



from ..collection import StaticScopeCollection, _url_scope


class _ComputeScopes(StaticScopeCollection):
    # The Compute service breaks the scopes/resource server convention: its resource
    # server is a service name and its scopes are built around the client ID.
    resource_server = "funcx_service"
    client_id = "facd7ccc-c5f4-42aa-916b-a0e270e2c2a9"

    all = _url_scope(client_id, "all")


ComputeScopes = _ComputeScopes()



import datetime

import globus_sdk
from globus_sdk.experimental.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# Globus Tutorial Collection 1
# https://app.globus.org/file-manager/collections/6c54cade-bde5-45c1-bdea-f4bd71dba2cc
SRC_COLLECTION = "6c54cade-bde5-45c1-bdea-f4bd71dba2cc"
SRC_PATH = "/home/share/godata/file1.txt"

# Globus Tutorial Collection 2
# https://app.globus.org/file-manager/collections/31ce9ba0-176d-45a5-add3-f37d233ba47d
DST_COLLECTION = "31ce9ba0-176d-45a5-add3-f37d233ba47d"
DST_PATH = "/~/example-timer-destination.txt"


def uses_data_access(client: globus_sdk.TransferClient, collection_id: str) -> bool:
    """
    Lookup the given collection ID.
    Having looked up the record, return `True` if it uses a `data_access` scope
    and `False` otherwise.
    """
    doc = client.get_endpoint(collection_id)
    if doc["entity_type"] != "GCSv5_mapped_collection":
        return False
    if doc["high_assurance"]:
        return False
    return True


# we'll define the timer as one which runs every hour for 3 days
# declare these data in the form of a "schedule" for the timer
#
# a wide variety of schedules are possible here; to setup a recurring timer:
# - you MUST declare an interval for the timer (`interval_seconds`)
# - you MAY declare an end condition (`end`)
schedule = globus_sdk.RecurringTimerSchedule(
    interval_seconds=3600,
    end={
        "condition": "time",
        "datetime": datetime.datetime.now() + datetime.timedelta(days=3),
    },
)


with UserApp("manage-timers-example", client_id=NATIVE_CLIENT_ID) as app:
    # as with an immediate data transfer, we take our input data and wrap them in
    # a TransferData object, representing the transfer task
    transfer_request = globus_sdk.TransferData(SRC_COLLECTION, DST_COLLECTION)
    transfer_request.add_item(SRC_PATH, DST_PATH)

    with (
        # we need a TransferClient, for the needs_data_access helper
        globus_sdk.TransferClient(app=app) as transfer_client,
        # create a TimersClient to interact with the service, and register any
        # data_access scopes for the collections
        globus_sdk.TimersClient(app=app) as timers_client,
    ):
        # Detect on each collection whether or not we need `data_access` scopes and
        # apply if necessary
        if uses_data_access(transfer_client, SRC_COLLECTION):
            timers_client.add_app_transfer_data_access_scope(SRC_COLLECTION)
        if uses_data_access(transfer_client, DST_COLLECTION):
            timers_client.add_app_transfer_data_access_scope(DST_COLLECTION)

        # submit the creation request to the service, printing out the
        # ID of your new timer after it's created -- you can find it in
        # https://app.globus.org/activity/timers
        timer = timers_client.create_timer(
            globus_sdk.TransferTimer(
                name=(
                    "create-timer-example "
                    f"[created at {datetime.datetime.now().isoformat()}]"
                ),
                body=transfer_request,
                schedule=schedule,
            )
        )
    print("Finished submitting timer.")
    print(f"timer_id: {timer['timer']['job_id']}")



from __future__ import annotations

import datetime
import typing as t
import uuid

from responses import matchers

from globus_sdk.testing.models import RegisteredResponse, ResponseList, ResponseSet

OWNER_URN = "urn:globus:auth:identity:a1234567-1234-1234-1234-123456789abc"


def generate_registered_api_summary(n: int) -> dict[str, t.Any]:
    """
    Generate a summary registered API object for list responses.

    :param n: The index number used to generate unique IDs and timestamps
    """
    api_id = str(uuid.UUID(int=n))
    base_time = datetime.datetime.fromisoformat("2024-01-01T00:00:00+00:00")
    created_timestamp = base_time + datetime.timedelta(days=n)
    updated_timestamp = created_timestamp + datetime.timedelta(hours=n)

    return {
        "id": api_id,
        "name": f"registered-api-{n}",
        "description": f"Test registered API number {n}",
        "created_timestamp": created_timestamp.isoformat(),
        "updated_timestamp": updated_timestamp.isoformat(),
    }


FIRST_REGISTERED_API_ID = str(uuid.UUID(int=0))

RESPONSES = ResponseSet(
    metadata={"first_registered_api_id": FIRST_REGISTERED_API_ID},
    default=RegisteredResponse(
        service="flows",
        path="/registered_apis",
        json={
            "registered_apis": [generate_registered_api_summary(0)],
            "limit": 1,
            "has_next_page": False,
            "marker": None,
        },
    ),
    paginated=ResponseList(
        RegisteredResponse(
            service="flows",
            path="/registered_apis",
            json={
                "registered_apis": [
                    generate_registered_api_summary(i) for i in range(10)
                ],
                "limit": 10,
                "has_next_page": True,
                "marker": "fake_marker_0",
            },
        ),
        RegisteredResponse(
            service="flows",
            path="/registered_apis",
            json={
                "registered_apis": [
                    generate_registered_api_summary(i) for i in range(10, 20)
                ],
                "limit": 10,
                "has_next_page": True,
                "marker": "fake_marker_1",
            },
            match=[matchers.query_param_matcher({"marker": "fake_marker_0"})],
        ),
        RegisteredResponse(
            service="flows",
            path="/registered_apis",
            json={
                "registered_apis": [
                    generate_registered_api_summary(i) for i in range(20, 25)
                ],
                "limit": 5,
                "has_next_page": False,
                "marker": None,
            },
            match=[matchers.query_param_matcher({"marker": "fake_marker_1"})],
        ),
        metadata={
            "owner_urn": OWNER_URN,
            "num_pages": 3,
            "expect_markers": ["fake_marker_0", "fake_marker_1", None],
            "total_items": 25,
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import TASK_GROUP_ID, TASK_ID, TASK_ID_2

TASK_BATCH_DOC = {
    "taskgroup_id": TASK_GROUP_ID,
    "create_websockets_queue": True,
    "tasks": [
        {"id": TASK_ID, "created_at": "2021-05-05T15:00:00.000000"},
        {"id": TASK_ID_2, "created_at": "2021-05-05T15:01:00.000000"},
    ],
}

RESPONSES = ResponseSet(
    metadata={
        "task_group_id": TASK_GROUP_ID,
        "task_id": TASK_ID,
        "task_id_2": TASK_ID_2,
    },
    default=RegisteredResponse(
        service="compute",
        path=f"/v2/taskgroup/{TASK_GROUP_ID}",
        method="GET",
        json=TASK_BATCH_DOC,
    ),
)



from __future__ import annotations

import uuid

from globus_sdk._payload import GlobusPayload


class DependentScopeSpec(GlobusPayload):
    """
    Utility class for creating dependent scope values as parameters to
    :meth:`AuthClient.create_scope <globus_sdk.AuthClient.create_scope>`
    and
    :meth:`AuthClient.update_scope <globus_sdk.AuthClient.update_scope>`.

    :param scope: The ID of the dependent scope
    :param optional: Whether or not the user can decline this specific scope without
        declining the whole consent.
    :param requires_refresh_token: Whether or not the dependency requires a refresh
        token.
    """

    def __init__(
        self,
        scope: uuid.UUID | str,
        optional: bool,
        requires_refresh_token: bool,
    ) -> None:
        super().__init__()
        self["scope"] = str(scope)
        self["optional"] = optional
        self["requires_refresh_token"] = requires_refresh_token



from unittest import mock

import pytest
import responses

import globus_sdk
from globus_sdk.testing import get_last_request
from tests.common import register_api_route_fixture_file

TASK1_ID = "b8872740-7edc-11ec-9f33-ed182a728dff"


@pytest.mark.parametrize(
    "add_kwargs",
    [
        {"timeout": 0, "polling_interval": 0},
        {"timeout": 0.5, "polling_interval": 0.5},
        {"timeout": 5, "polling_interval": 0.5},
        {"timeout": -5, "polling_interval": 5},
    ],
)
def test_task_wait_bad_args_min_wait(client, mocksleep, add_kwargs):
    # register task mock data even though it should not be needed
    register_api_route_fixture_file(
        "transfer", f"/task/{TASK1_ID}", "get_task1_active.json"
    )

    with pytest.raises(globus_sdk.GlobusSDKUsageError):
        client.task_wait(TASK1_ID, **add_kwargs)

    # no requests sent, no sleep done
    assert get_last_request() is None
    mocksleep.assert_not_called()


def test_task_wait_success_case(client, mocksleep):
    # first the task will show as active, then as succeeded
    register_api_route_fixture_file(
        "transfer", f"/task/{TASK1_ID}", "get_task1_active.json"
    )
    register_api_route_fixture_file(
        "transfer", f"/task/{TASK1_ID}", "get_task1_succeeded.json"
    )

    # do the task wait, it should return true (the task completed in time)
    result = client.task_wait(TASK1_ID, timeout=5, polling_interval=1)
    assert result is True

    # one sleep, two network calls
    mocksleep.assert_called_once_with(1)
    assert len(responses.calls) == 2


def test_task_wait_unfinished_case(client, mocksleep):
    # the task is in the active state no matter how many times we ask
    register_api_route_fixture_file(
        "transfer", f"/task/{TASK1_ID}", "get_task1_active.json"
    )

    # do the task wait, it should return false (the task didn't complete)
    result = client.task_wait(TASK1_ID, timeout=5, polling_interval=1)
    assert result is False

    # a number of sleeps equal to timeout/polling_interval
    # a number of requests equal to sleeps+1
    mocksleep.assert_has_calls([mock.call(1) for x in range(5)])
    assert len(responses.calls) == 6



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="transfer",
        method="POST",
        path=f"/v0.10/operation/endpoint/{ENDPOINT_ID}/mkdir",
        json={
            "DATA_TYPE": "mkdir_result",
            "code": "DirectoryCreated",
            "message": "The directory was created successfully",
            "request_id": "ShbIUzrWT",
            "resource": f"/v0.10/operation/endpoint/{ENDPOINT_ID}/mkdir",
        },
    ),
)



from .errors import LocalServerEnvironmentalLoginError, LocalServerLoginError
from .local_server_login_flow_manager import LocalServerLoginFlowManager

__all__ = [
    "LocalServerLoginError",
    "LocalServerEnvironmentalLoginError",
    "LocalServerLoginFlowManager",
]



import pytest

from globus_sdk import exc
from globus_sdk.token_storage import SQLiteTokenStorage
from globus_sdk.token_storage.legacy import SQLiteAdapter


@pytest.fixture
def db_file(tmp_path):
    return tmp_path / "test.db"


@pytest.fixture
def adapters_to_close():
    data = set()
    yield data
    for x in data:
        x.close()


@pytest.fixture
def make_adapter(adapters_to_close, db_file):
    def func(*args, **kwargs):
        if len(args) == 0 and "filepath" not in kwargs:
            args = (db_file,)
        ret = SQLiteTokenStorage(*args, **kwargs)
        adapters_to_close.add(ret)
        return ret

    return func


@pytest.mark.parametrize("kwargs", [{}, {"namespace": "foo"}])
def test_constructor(kwargs, db_file, make_adapter):
    make_adapter(db_file, **kwargs)
    assert db_file.exists()


def test_constructor_rejects_memory_db(make_adapter):
    with pytest.raises(
        exc.GlobusSDKUsageError,
        match="SQLiteTokenStorage cannot be used with a ':memory:' database",
    ):
        make_adapter(":memory:")


def test_store_and_get_token_data_by_resource_server(
    mock_token_data_by_resource_server, make_adapter
):
    adapter = make_adapter()
    adapter.store_token_data_by_resource_server(mock_token_data_by_resource_server)

    gotten = adapter.get_token_data_by_resource_server()

    for resource_server in ["resource_server_1", "resource_server_2"]:
        assert (
            mock_token_data_by_resource_server[resource_server].to_dict()
            == gotten[resource_server].to_dict()
        )


def test_multiple_adapters_store_and_retrieve(mock_response, db_file, make_adapter):
    adapter1 = make_adapter(db_file)
    adapter2 = make_adapter(db_file)
    adapter1.store_token_response(mock_response)

    assert adapter2.get_token_data("resource_server_1").access_token == "access_token_1"
    assert adapter2.get_token_data("resource_server_2").access_token == "access_token_2"


def test_multiple_adapters_store_and_retrieve_different_namespaces(
    mock_response, db_file, make_adapter
):
    adapter1 = make_adapter(db_file, namespace="foo")
    adapter2 = make_adapter(db_file, namespace="bar")
    adapter1.store_token_response(mock_response)

    data = adapter2.get_token_data_by_resource_server()
    assert data == {}


def test_load_missing_token_data(make_adapter):
    adapter = make_adapter()
    assert adapter.get_token_data_by_resource_server() == {}
    assert adapter.get_token_data("resource_server_1") is None


def test_remove_tokens(mock_response, make_adapter):
    adapter = make_adapter()
    adapter.store_token_response(mock_response)

    removed = adapter.remove_token_data("resource_server_1")
    assert removed
    assert adapter.get_token_data("resource_server_1") is None
    assert adapter.get_token_data("resource_server_2").access_token == "access_token_2"

    removed = adapter.remove_token_data("resource_server_1")
    assert not removed


def test_iter_namespaces(mock_response, db_file, make_adapter):
    foo_adapter = make_adapter(db_file, namespace="foo")
    bar_adapter = make_adapter(db_file, namespace="bar")

    for adapter in [foo_adapter, bar_adapter]:
        assert list(adapter.iter_namespaces()) == []

    foo_adapter.store_token_response(mock_response)

    for adapter in [foo_adapter, bar_adapter]:
        assert list(adapter.iter_namespaces()) == ["foo"]

    bar_adapter.store_token_response(mock_response)

    for adapter in [foo_adapter, bar_adapter]:
        assert set(adapter.iter_namespaces()) == {"foo", "bar"}


def test_migrate_from_v1_adapter(mock_response, db_file, make_adapter):
    # store data with SQLiteAdapter
    old_adapter = SQLiteAdapter(db_file)
    old_adapter.store(mock_response)
    old_adapter.close()

    # retrieve data with SQLiteTokenStorage using the same file
    new_adapter = make_adapter(db_file)
    assert (
        new_adapter.get_token_data("resource_server_1").access_token == "access_token_1"
    )
    assert (
        new_adapter.get_token_data("resource_server_2").access_token == "access_token_2"
    )



import json

import pytest


def test_expand_testing_fixture_fails_on_bad_reference(sphinx_runner, capsys):
    sphinx_runner.ensure_failure(
        ".. expandtestfixture:: NO_SUCH_FIXTURE",
    )

    captured = capsys.readouterr()
    err_lines = captured.err.splitlines()

    test_line = None
    for line in err_lines:
        if "ValueError: no fixtures defined" in line:
            test_line = line
            break
    else:
        pytest.fail("Didn't find 'ValueError: no fixtures defined' in stderr")

    assert (
        "no fixtures defined for globus_sdk.testing.data.NO_SUCH_FIXTURE" in test_line
    )


# choose an arbitrary example fixture to test
def test_expand_testing_fixture_on_valid_fixture(sphinx_runner):
    etree = sphinx_runner.to_etree(
        ".. expandtestfixture:: groups.set_group_policies",
    )

    code_block = etree.find("./literal_block")
    assert code_block is not None
    assert code_block.get("language") == "json"

    # check against the known values for this fixture
    data = json.loads(code_block.text)
    assert data["is_high_assurance"] is False
    assert data["group_visibility"] == "private"


# choose an arbitrary example fixture to test with multiple cases
def test_expand_testing_fixture_on_non_default_case(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. expandtestfixture:: auth.userinfo
            :case: unauthorized
        """,
    )

    code_block = etree.find("./literal_block")
    assert code_block is not None
    assert code_block.get("language") == "json"

    # check against the known values for this fixture
    data = json.loads(code_block.text)
    assert data["error_description"] == "Unauthorized"
    assert data["errors"][0]["status"] == "401"



import time
from unittest import mock

import pytest


@pytest.fixture
def mock_response():
    res = mock.Mock()
    expiration_time = int(time.time()) + 3600
    res.by_resource_server = {
        "resource_server_1": {
            "access_token": "access_token_1",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_1",
            "resource_server": "resource_server_1",
            "scope": "scope1",
            "token_type": "bearer",
        },
        "resource_server_2": {
            "access_token": "access_token_2",
            "expires_in": expiration_time,
            "refresh_token": "refresh_token_2",
            "resource_server": "resource_server_2",
            "scope": "scope2 scope2:0 scope2:1",
            "token_type": "bearer",
        },
    }
    return res


@pytest.fixture
def mock_refresh_response():
    res = mock.Mock()
    expiration_time = int(time.time()) + 3600
    res.by_resource_server = {
        "resource_server_2": {
            "access_token": "access_token_2_refreshed",
            "expires_in": expiration_time,
            "refresh_token": "refresh_token_2",
            "resource_server": "resource_server_2",
            "scope": "scope2 scope2:0 scope2:1",
            "token_type": "bearer",
        }
    }
    return res



import pytest

import globus_sdk
from globus_sdk.scopes import Scope


def test_specific_flow_client_class_errors_on_scope_access():
    scopes = globus_sdk.SpecificFlowClient.scopes
    assert scopes is not None

    # for the 'user' scope, which is well-defined, we get a special error
    with pytest.raises(AttributeError) as excinfo:
        scopes.user

    err = excinfo.value
    assert (
        "It is not valid to attempt to access the 'scopes' attribute of the "
        "SpecificFlowClient class."
    ) in str(err)

    # but for any other scope we get something a little more generic
    with pytest.raises(AttributeError) as excinfo:
        scopes.demuddle

    err = excinfo.value
    assert str(err).endswith("has no attribute 'demuddle'")


def test_specific_flow_client_class_errors_on_resource_server_access():
    scopes = globus_sdk.SpecificFlowClient.scopes
    assert scopes is not None

    # access via the scopes object raises an error
    with pytest.raises(AttributeError) as excinfo:
        scopes.resource_server

    err = excinfo.value
    assert (
        "It is not valid to attempt to access the 'resource_server' attribute of the "
        "SpecificFlowClient class."
    ) in str(err)

    # and access via the client class raises the same error
    with pytest.raises(AttributeError) as excinfo:
        globus_sdk.SpecificFlowClient.resource_server

    err = excinfo.value
    assert (
        "It is not valid to attempt to access the 'resource_server' attribute of the "
        "SpecificFlowClient class."
    ) in str(err)


def test_specific_flow_client_instance_supports_scope_access():
    client = globus_sdk.SpecificFlowClient("foo")
    scopes = client.scopes
    assert scopes is not None

    # for the 'user' scope, we get a string
    user_scope = scopes.user
    assert isinstance(user_scope, Scope)
    assert str(user_scope).endswith("flow_foo_user")

    # but for any other scope we still get the generic attribute error
    with pytest.raises(AttributeError) as excinfo:
        scopes.demuddle

    err = excinfo.value
    assert str(err).endswith("has no attribute 'demuddle'")


def test_specific_flow_client_instance_supports_resource_server_access():
    client = globus_sdk.SpecificFlowClient("foo")
    resource_server = client.resource_server
    assert isinstance(resource_server, str)
    assert resource_server == "foo"



from .base import FileAdapter, StorageAdapter
from .file_adapters import SimpleJSONFileAdapter
from .memory_adapter import MemoryAdapter
from .sqlite_adapter import SQLiteAdapter

__all__ = (
    "StorageAdapter",
    "FileAdapter",
    "SimpleJSONFileAdapter",
    "SQLiteAdapter",
    "MemoryAdapter",
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import FUNCTION_CODE, FUNCTION_ID, FUNCTION_NAME

DEFAULT_RESPONSE_DOC = {
    "function_uuid": FUNCTION_ID,
}

RESPONSES = ResponseSet(
    metadata={
        "function_id": FUNCTION_ID,
        "function_name": FUNCTION_NAME,
        "function_code": FUNCTION_CODE,
    },
    default=RegisteredResponse(
        service="compute",
        path="/v3/functions",
        method="POST",
        json=DEFAULT_RESPONSE_DOC,
    ),
)



from __future__ import annotations

import logging
import sys
import typing as t

from globus_sdk import exc

from ._auth_requirements_error import GARE
from ._variants import (
    LegacyAuthorizationParametersError,
    LegacyAuthRequirementsErrorVariant,
    LegacyConsentRequiredAPError,
    LegacyConsentRequiredTransferError,
    LegacyDependentConsentRequiredAuthError,
)

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


AnyErrorDocumentType: TypeAlias = (
    "exc.GlobusAPIError | exc.ErrorSubdocument | dict[str, t.Any]"
)

log = logging.getLogger(__name__)


def to_gare(error: AnyErrorDocumentType) -> GARE | None:
    """
    Converts a GlobusAPIError, ErrorSubdocument, or dict into a
    GARE by attempting to match to GARE (preferred) or legacy variants.

    .. note::

        A GlobusAPIError may contain multiple errors, and in this case only a single
        GARE is returned for the first error that matches a known format.


    If the provided error does not match a known format, None is returned.

    :param error: The error to convert.
    """
    from globus_sdk.exc import ErrorSubdocument, GlobusAPIError

    # GlobusAPIErrors may contain more than one error, so we consider all of them
    # even though we only return the first.
    if isinstance(error, GlobusAPIError):
        # first, try to parse a GARE from the root document,
        # and if we can do so, return it
        authreq_error = _lenient_dict2gare(error.raw_json)
        if authreq_error is not None:
            return authreq_error

        # Iterate over ErrorSubdocuments
        for subdoc in error.errors:
            authreq_error = _lenient_dict2gare(subdoc.raw)
            if authreq_error is not None:
                # Return only the first auth requirements error we encounter
                return authreq_error

        # We failed to find a Globus Auth Requirements Error
        return None
    elif isinstance(error, ErrorSubdocument):
        return _lenient_dict2gare(error.raw)
    else:
        return _lenient_dict2gare(error)


def to_gares(errors: list[AnyErrorDocumentType]) -> list[GARE]:
    """
    Converts a list of GlobusAPIErrors, ErrorSubdocuments, or dicts into a list of
    GAREs by attempting to match each error to GARE (preferred) or legacy variants.

    .. note::

        A GlobusAPIError may contain multiple errors, so the result
        list could be longer than the provided list.

    If no errors match any known formats, an empty list is returned.

    :param errors: The errors to convert.
    """
    maybe_gares: list[GARE | None] = []
    for error in errors:
        # when handling an API error, avoid `to_gare(error)` because that will
        # only unpack a single result
        if isinstance(error, exc.GlobusAPIError):
            # Use the ErrorSubdocuments when handling API error types
            maybe_gares.extend(to_gare(e) for e in error.errors)
            # Also use the root document, but only if there is an `"errors"`
            # key inside of the error document
            # Why? Because the *default* for `.errors` when there is no inner
            # `"errors"` array is an array containing the root document as a
            # subdocument
            if isinstance(error.raw_json, dict) and "errors" in error.raw_json:
                # use dict parsing directly so that the native descent in 'to_gare'
                # to subdocuments does not apply in this case
                maybe_gares.append(_lenient_dict2gare(error.raw_json))
        else:
            maybe_gares.append(to_gare(error))

    # Remove any errors that did not resolve to a Globus Auth Requirements Error
    return [error for error in maybe_gares if error is not None]


def _lenient_dict2gare(error_dict: dict[str, t.Any] | None) -> GARE | None:
    """
    Parse a GARE from a dict, accepting legacy variants.

    If given ``None``, returns ``None``. This allows this to accept inputs
    which are themselves dict|None.

    :param error_dict: the error input
    :eturns: ``None`` on a failed parse
    """
    if error_dict is None:
        return None

    # Prefer a proper auth requirements error, if possible
    try:
        return GARE.from_dict(error_dict)
    except exc.ValidationError as err:
        log.debug(f"Failed to parse error as 'GARE' ({err})")

    supported_variants: list[type[LegacyAuthRequirementsErrorVariant]] = [
        LegacyAuthorizationParametersError,
        LegacyConsentRequiredTransferError,
        LegacyConsentRequiredAPError,
        LegacyDependentConsentRequiredAuthError,
    ]
    for variant in supported_variants:
        try:
            return variant.from_dict(error_dict).to_auth_requirements_error()
        except exc.ValidationError as err:
            log.debug(f"Failed to parse error as '{variant.__name__}' ({err})")

    return None


def is_gare(error: AnyErrorDocumentType) -> bool:
    """
    Return True if the provided error matches a known
    Globus Auth Requirements Error format.

    :param error: The error to check.
    """
    return to_gare(error) is not None


def has_gares(errors: list[AnyErrorDocumentType]) -> bool:
    """
    Return True if any of the provided errors match a known
    Globus Auth Requirements Error format.

    :param errors: The errors to check.
    """
    return any(is_gare(error) for error in errors)



#!/usr/bin/env python

import argparse
import datetime

import globus_sdk
from globus_sdk.experimental.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
USER_APP = UserApp("manage-timers-example", client_id=NATIVE_CLIENT_ID)


def main():
    parser = argparse.ArgumentParser()
    # the source, destination, and path to a file or dir to sync
    parser.add_argument("SOURCE_COLLECTION")
    parser.add_argument("DESTINATION_COLLECTION")
    parser.add_argument("PATH")
    parser.add_argument(
        "--interval-seconds",
        help="How frequently the timer runs, in seconds (default: 1 hour)",
        default=3600,
        type=int,
    )
    parser.add_argument(
        "--days",
        help="How many days to run the timer (default: 2)",
        default=2,
        type=int,
    )
    args = parser.parse_args()

    client = globus_sdk.TimersClient(app=USER_APP)

    body = globus_sdk.TransferData(
        source_endpoint=args.SOURCE_COLLECTION,
        destination_endpoint=args.DESTINATION_COLLECTION,
    )
    body.add_item(args.PATH, args.PATH)

    # the timer will run until the end date, on whatever interval was requested
    schedule = globus_sdk.RecurringTimerSchedule(
        interval_seconds=args.interval_seconds,
        end={
            "condition": "time",
            "datetime": datetime.datetime.now() + datetime.timedelta(days=args.days),
        },
    )

    timer = client.create_timer(
        timer=globus_sdk.TransferTimer(
            name=(
                "create-timer-example "
                f"[created at {datetime.datetime.now().isoformat()}]"
            ),
            body=body,
            schedule=schedule,
        )
    )
    print("Finished submitting timer.")
    print(f"timer_id: {timer['timer']['job_id']}")


if __name__ == "__main__":
    main()



class GlobusError(Exception):
    """
    Root of the Globus Exception hierarchy.
    """


class GlobusSDKUsageError(GlobusError, ValueError):
    """
    A ``GlobusSDKUsageError`` may be thrown in cases in which the SDK
    detects that it is being used improperly.

    These errors typically indicate that some contract regarding SDK usage
    (e.g. required order of operations) has been violated.
    """


class ValidationError(GlobusError, ValueError):
    """
    A ``ValidationError`` may be raised when the SDK is instructed to handle or
    parse data which can be seen to be invalid without an external service
    interaction.

    These errors typically do not indicate a usage error similar to
    ``GlobusSDKUsageError``, but rather that the data is invalid.
    """



from __future__ import annotations

import typing as t

import requests

from globus_sdk import exc


class GCSAPIError(exc.GlobusAPIError):
    """
    Error class for the GCS Manager API client
    """

    def __init__(self, r: requests.Response) -> None:
        self.detail_data_type: str | None = None
        self.detail: None | str | dict[str, t.Any] = None
        super().__init__(r)

    def _get_args(self) -> list[t.Any]:
        args = super()._get_args()
        args.append(self.detail_data_type)
        # only add detail if it's a string (don't want to put a large object into
        # stacktraces)
        if isinstance(self.detail, str):
            args.append(self.detail)
        return args

    def _post_parse_hook(self) -> bool:
        # detail can be a full document, so fetch, then look for a DATA_TYPE
        # and expose it as a top-level attribute for easy access
        self.detail = self._dict_data.get("detail")
        if isinstance(self.detail, dict) and "DATA_TYPE" in self.detail:
            self.detail_data_type = self.detail["DATA_TYPE"]
        return True



import globus_sdk

# this is the tutorial client ID
# replace this string with your ID for production use
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# create a client for interactions with Globus Auth
auth_client = globus_sdk.NativeAppAuthClient(CLIENT_ID)

# using that client, do a login flow for Globus Groups credentials
auth_client.oauth2_start_flow(
    requested_scopes=globus_sdk.GroupsClient.scopes.view_my_groups_and_memberships
)
authorize_url = auth_client.oauth2_get_authorize_url()
print(f"Please go to this URL and login:\n\n{authorize_url}\n")

auth_code = input("Please enter the code here: ").strip()
tokens = auth_client.oauth2_exchange_code_for_tokens(auth_code)

# extract tokens from the response which match Globus Groups
groups_tokens = tokens.by_resource_server[globus_sdk.GroupsClient.resource_server]

# construct an AccessTokenAuthorizer and use it to construct the GroupsClient
groups_client = globus_sdk.GroupsClient(
    authorizer=globus_sdk.AccessTokenAuthorizer(groups_tokens["access_token"])
)

# call out to the Groups service to get a listing
my_groups = groups_client.get_my_groups()

# print in CSV format
print("ID,Name,Roles")
for group in my_groups:
    roles = "|".join({m["role"] for m in group["my_memberships"]})
    print(",".join([group["id"], f'"{group["name"]}"', roles]))



from __future__ import annotations

import inspect
import sys
import typing as t

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Serializable:
    """
    This is a base class for helpers which represent data which can be
    loaded to or from a dictionary.

    Serializable classes:
    - know what fields they have, based on their initializer signatures
    - support `to_dict()` and `from_dict()` conversions
    - typically use `globus_sdk._internal.guards.validators` to check attribute types
    """

    _EXCLUDE_VARS: t.ClassVar[tuple[str, ...]] = ("self", "extra")
    extra: dict[str, t.Any]

    @classmethod
    def _supported_fields(cls) -> list[str]:
        signature = inspect.signature(cls.__init__)
        return [
            name
            for name in signature.parameters.keys()
            if name not in cls._EXCLUDE_VARS
        ]

    @classmethod
    def from_dict(cls, data: dict[str, t.Any]) -> Self:
        """
        Instantiate from a dictionary.

        :param data: The dictionary to create the error from.
        """

        # Extract any extra fields
        extras = {k: v for k, v in data.items() if k not in cls._supported_fields()}
        kwargs: dict[str, t.Any] = {"extra": extras}
        # Ensure required fields are supplied
        for field_name in cls._supported_fields():
            kwargs[field_name] = data.get(field_name)

        return cls(**kwargs)

    def to_dict(self, include_extra: bool = False) -> dict[str, t.Any]:
        """
        Render to a dictionary.

        :param include_extra: Whether to include stored extra (non-standard) fields in
            the returned dictionary.
        """
        result = {}

        # Set any authorization parameters
        for field in self._supported_fields():
            value = getattr(self, field)
            if value is not None:
                if isinstance(value, Serializable):
                    value = value.to_dict(include_extra=include_extra)
                result[field] = value

        # Set any extra fields
        if include_extra:
            result.update(self.extra)

        return result



from __future__ import annotations

import json
import logging
import textwrap
import time
import typing as t

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from globus_sdk import exc
from globus_sdk.response import GlobusHTTPResponse

from .._common import SupportsJWKMethods
from ..id_token_decoder import IDTokenDecoder

log = logging.getLogger(__name__)


def _convert_token_info_dict(
    source_dict: GlobusHTTPResponse,
) -> dict[str, t.Any]:
    """
    Extract a set of fields into a new dict for indexing by resource server.
    Allow for these fields to be `None` when absent:
        - "refresh_token"
        - "token_type"
    """
    expires_in = source_dict.get("expires_in", 0)

    return {
        "scope": source_dict["scope"],
        "access_token": source_dict["access_token"],
        "refresh_token": source_dict.get("refresh_token"),
        "token_type": source_dict.get("token_type"),
        "expires_at_seconds": int(time.time() + expires_in),
        "resource_server": source_dict["resource_server"],
    }


class _ByScopesGetter:
    """
    A fancy dict-like object for looking up token data by scope name.
    Allows usage like

    >>> tokens = OAuthTokenResponse(...)
    >>> tok = tokens.by_scopes['openid profile']['access_token']
    """

    def __init__(self, scope_map: dict[str, t.Any]) -> None:
        self.scope_map = scope_map

    def __str__(self) -> str:
        return json.dumps(self.scope_map)

    def __iter__(self) -> t.Iterator[str]:
        """iteration gets you every individual scope"""
        return iter(self.scope_map.keys())

    def __getitem__(self, scopename: str) -> dict[str, str | int]:
        if not isinstance(scopename, str):
            raise KeyError(f'by_scopes cannot contain non-string value "{scopename}"')

        # split on spaces
        scopes = scopename.split()
        # collect every matching token in a set to dedup
        # but collect actual results (dicts) in a list
        rs_names = set()
        toks = []
        for scope in scopes:
            try:
                rs_names.add(self.scope_map[scope]["resource_server"])
                toks.append(self.scope_map[scope])
            except KeyError as err:
                raise KeyError(
                    (
                        'Scope specifier "{}" contains scope "{}" '
                        "which was not found"
                    ).format(scopename, scope)
                ) from err
        # if there isn't exactly 1 token, it's an error
        if len(rs_names) != 1:
            raise KeyError(
                'Scope specifier "{}" did not match exactly one token!'.format(
                    scopename
                )
            )
        # pop the only element in the set
        return t.cast(t.Dict[str, t.Union[str, int]], toks.pop())

    def __contains__(self, item: str) -> bool:
        """
        contains is driven by checking against getitem
        that way, the definitions are always "in sync" if we update them in
        the future
        """
        try:
            self.__getitem__(item)
            return True
        except KeyError:
            pass

        return False


class OAuthTokenResponse(GlobusHTTPResponse):
    """
    Class for responses from the OAuth2 code for tokens exchange used in
    3-legged OAuth flows.
    """

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._init_rs_dict()
        self._init_scopes_getter()

    def _init_scopes_getter(self) -> None:
        scope_map = {}
        for _rs, tok_data in self._by_resource_server.items():
            for s in tok_data["scope"].split():
                scope_map[s] = tok_data
        self._by_scopes = _ByScopesGetter(scope_map)

    def _init_rs_dict(self) -> None:
        # call the helper at the top level
        self._by_resource_server = {
            self["resource_server"]: _convert_token_info_dict(self)
        }
        # call the helper on everything in 'other_tokens'
        self._by_resource_server.update(
            {
                unprocessed_item["resource_server"]: _convert_token_info_dict(
                    unprocessed_item
                )
                for unprocessed_item in self["other_tokens"]
            }
        )

    @property
    def by_resource_server(self) -> dict[str, dict[str, t.Any]]:
        """
        Representation of the token response in a ``dict`` indexed by resource
        server.

        Although ``OAuthTokenResponse.data`` is still available and
        valid, this representation is typically more desirable for applications
        doing inspection of access tokens and refresh tokens.
        """
        return self._by_resource_server

    @property
    def by_scopes(self) -> _ByScopesGetter:
        """
        Representation of the token response in a dict-like object indexed by
        scope name (or even space delimited scope names, so long as they match
        the same token).

        If you request scopes `scope1 scope2 scope3`, where `scope1` and
        `scope2` are for the same service (and therefore map to the same
        token), but `scope3` is for a different service, the following forms of
        access are valid:

        >>> tokens = ...
        >>> # single scope
        >>> token_data = tokens.by_scopes['scope1']
        >>> token_data = tokens.by_scopes['scope2']
        >>> token_data = tokens.by_scopes['scope3']
        >>> # matching scopes
        >>> token_data = tokens.by_scopes['scope1 scope2']
        >>> token_data = tokens.by_scopes['scope2 scope1']
        """
        return self._by_scopes

    def decode_id_token(
        self,
        openid_configuration: None | GlobusHTTPResponse | dict[str, t.Any] = None,
        jwk: RSAPublicKey | None = None,
        jwt_params: dict[str, t.Any] | None = None,
    ) -> dict[str, t.Any]:
        """
        Parse the included ID Token (OIDC) as a dict and return it.

        If you provide the `jwk`, you must also provide `openid_configuration`.

        :param openid_configuration: The OIDC config as a GlobusHTTPResponse or dict.
            When not provided, it will be fetched automatically.
        :param jwk: The JWK as a cryptography public key object. When not provided, it
            will be fetched and parsed automatically.
        :param jwt_params: An optional dict of parameters to pass to the jwt decode
            step. If ``"leeway"`` is included, it will be passed as the ``leeway``
            parameter, and all other values are passed as ``options``.
        """
        id_token = self["id_token"]
        log.debug('Decoding ID Token "%s"', id_token)
        if not isinstance(self.client, SupportsJWKMethods):
            raise exc.GlobusSDKUsageError(
                "decode_id_token() requires a client which supports JWK methods. "
                "This error suggests that an improper client type is attached to "
                "the token response."
            )
        else:
            auth_client: SupportsJWKMethods = self.client

        decoder = IDTokenDecoder(auth_client)

        if openid_configuration:
            decoder.store_openid_configuration(openid_configuration)
        else:
            if jwk:
                raise exc.GlobusSDKUsageError(
                    "passing jwk without openid configuration is not allowed"
                )
        if jwk:
            decoder.store_jwk(jwk)

        jwt_params = jwt_params or {}
        if "leeway" in jwt_params:
            jwt_params = jwt_params.copy()
            decoder.jwt_leeway = jwt_params.pop("leeway")

        decoder.jwt_options = jwt_params

        return decoder.decode(id_token)

    def __str__(self) -> str:
        by_rs = json.dumps(self.by_resource_server, indent=2, separators=(",", ": "))
        id_token_to_print = t.cast(t.Optional[str], self.get("id_token"))
        if id_token_to_print is not None:
            id_token_to_print = id_token_to_print[:10] + "... (truncated)"
        return (
            f"{self.__class__.__name__}:\n"
            + f"  id_token: {id_token_to_print}\n"
            + "  by_resource_server:\n"
            + textwrap.indent(by_rs, "    ")
        )


class OAuthAuthorizationCodeResponse(OAuthTokenResponse):
    """
    Class for responses from the OAuth2 'authorization_code' grant.

    This class of response is returned by methods which get new tokens via a
    code-exchange, as in 3-legged OAuth or PKCE.

    For example,
    :meth:`globus_sdk.ConfidentialAppAuthClient.oauth2_exchange_code_for_tokens`
    will return an ``OAuthAuthorizationCodeResponse``.
    """


class OAuthClientCredentialsResponse(OAuthTokenResponse):
    """
    Class for responses from the OAuth2 'client_credentials' grant.

    This class of response is returned by methods which get new tokens by means of
    client credentials, namely
    :meth:`globus_sdk.ConfidentialAppAuthClient.oauth2_client_credentials_tokens`.
    """


class OAuthRefreshTokenResponse(OAuthTokenResponse):
    """
    Class for responses from the OAuth2 'refresh_token' grant.

    This class of response is returned by methods which get new tokens from an
    existing Refresh Token, e.g.,
    :meth:`globus_sdk.NativeAppAuthClient.oauth2_refresh_token`.
    """


class OAuthDependentTokenResponse(OAuthTokenResponse):
    """
    Class for responses from the OAuth2
    'urn:globus:auth:grant_type:dependent_token' grant.

    This is an extension grant type defined by Globus.

    :meth:`oauth2_get_dependent_tokens \
    <globus_sdk.ConfidentialAppAuthClient.oauth2_get_dependent_tokens>`
    provides this response, and includes some documentation on its proper usage.
    """

    def _init_rs_dict(self) -> None:
        # call the helper on everything in the response array
        self._by_resource_server = {
            unprocessed_item["resource_server"]: _convert_token_info_dict(
                unprocessed_item
            )
            for unprocessed_item in self.data
        }

    def decode_id_token(
        self,
        openid_configuration: None | (GlobusHTTPResponse | dict[str, t.Any]) = None,
        jwk: RSAPublicKey | None = None,
        jwt_params: dict[str, t.Any] | None = None,
    ) -> dict[str, t.Any]:
        # just in case
        raise NotImplementedError(
            "OAuthDependentTokenResponse.decode_id_token() is not and cannot "
            "be implemented. Dependent Tokens data does not include an "
            "id_token"
        )



from __future__ import annotations

import abc
import time
import typing as t

import globus_sdk
from globus_sdk.scopes.consents import ConsentForest

from ..token_data import TokenStorageData
from .context import TokenValidationContext
from .errors import (
    ExpiredTokenError,
    IdentityMismatchError,
    MissingIdentityError,
    MissingTokenError,
    UnmetScopeRequirementsError,
)


class TokenDataValidator(abc.ABC):
    """
    The abstract base class for custom token validation logic.

    Implementations should raise a :class:`TokenValidationError` if a validation error
    is encountered.
    """

    @abc.abstractmethod
    def before_store(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,
    ) -> None:
        """
        Validate token data against this validator's constraints before it is
        written to token storage.

        :param token_data_by_resource_server: The data to validate.
        :param context: The validation context object, containing state of the system at
            the time of validation.

        :raises TokenValidationError: On failure.
        """

    @abc.abstractmethod
    def after_retrieve(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,
    ) -> None:
        """
        Validate token data against this validator's constraints after it is
        retrieved from token storage.

        :param token_data_by_resource_server: The data to validate.
        :param context: The validation context object, containing state of the system at
            the time of validation.

        :raises TokenValidationError: On failure.
        """


class _OnlyBeforeValidator(TokenDataValidator, abc.ABC):
    def after_retrieve(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,
    ) -> None:
        return None


class _OnlyAfterValidator(TokenDataValidator, abc.ABC):
    def before_store(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,
    ) -> None:
        return None


class HasRefreshTokensValidator(_OnlyAfterValidator):
    """
    A validator to validate that token data contains refresh tokens.

    *  This validator only runs `after_retrieve`.

    :raises MissingTokenError: If any token data does not contain a refresh token.
    """

    def after_retrieve(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,  # pylint: disable=unused-argument
    ) -> None:
        for token_data in token_data_by_resource_server.values():
            if token_data.refresh_token is None:
                msg = f"No refresh_token for {token_data.resource_server}"
                raise MissingTokenError(msg, resource_server=token_data.resource_server)


class NotExpiredValidator(_OnlyAfterValidator):
    """
    A validator to validate that token data has not expired.

    * This validator only runs `after_retrieve`.

    :raises ExpiredTokenError: If any token data shows has expired.
    """

    def after_retrieve(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,  # pylint: disable=unused-argument
    ) -> None:
        for token_data in token_data_by_resource_server.values():
            if token_data.expires_at_seconds < time.time():
                raise ExpiredTokenError(token_data.expires_at_seconds)


class UnchangingIdentityIDValidator(_OnlyBeforeValidator):
    """
    A validator to validate that user identity does not change across token grants.

    *   This validator only runs `before_store`.

    :raises IdentityMismatchError: If identity info changes across token storage
        operations.
    :raises MissingIdentityError: If the token data did not have any identity
        information.
    """

    def before_store(
        self,
        token_data_by_resource_server: t.Mapping[  # pylint: disable=unused-argument
            str, TokenStorageData
        ],
        context: TokenValidationContext,
    ) -> None:
        if context.token_data_identity_id is None:
            raise MissingIdentityError(
                "Token grant response doesn't contain an id_token. This normally "
                "occurs if the auth flow didn't include 'openid' alongside other "
                "scopes."
            )

        # no prior ID means we cannot validate the content further
        if context.prior_identity_id is None:
            return

        if context.token_data_identity_id != context.prior_identity_id:
            raise IdentityMismatchError(
                "Detected a change in identity associated with the token data.",
                stored_id=context.prior_identity_id,
                new_id=context.token_data_identity_id,
            )


class ScopeRequirementsValidator(TokenDataValidator):
    """
    A validator to validate that token data meets scope requirements.

    A scope requirement, i.e., "transfer:all[collection:data_access]", can be broken
    into two parts: the **root scope**, "transfer:all", and one or more
    **dependent scopes**, "collection:data_access".

    *   On `before_store`, this validator only evaluates **root** scope requirements.
    *   On `after_retrieve`, this validator evaluates both **root** and **dependent**
        scope requirements.

    .. note::

        Dependent scopes are only evaluated if an identity ID is extractable from
        token grants. If no identity ID is available, dependent scope evaluation
        is silently skipped.

    :param scope_requirements: A mapping of resource servers to required scopes.
    :param consent_client: An AuthClient to fetch consents with. This auth client must
        have (or have access to) any valid Globus Auth scoped token.

    :raises UnmetScopeRequirementsError: If any scope requirements are not met.
    """

    def __init__(
        self,
        scope_requirements: t.Mapping[str, t.Sequence[globus_sdk.Scope]],
        consent_client: globus_sdk.AuthClient,
    ) -> None:
        self.scope_requirements = scope_requirements
        self.consent_client: globus_sdk.AuthClient = consent_client
        self._cached_consent_forest: ConsentForest | None = None

    def before_store(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,
    ) -> None:
        identity_id = context.token_data_identity_id or context.prior_identity_id
        for resource_server, token_data in token_data_by_resource_server.items():
            self._validate_token_data_meets_scope_requirements(
                resource_server=resource_server,
                token_data=token_data,
                identity_id=identity_id,
                eval_dependent=False,
            )

    def after_retrieve(
        self,
        token_data_by_resource_server: t.Mapping[str, TokenStorageData],
        context: TokenValidationContext,
    ) -> None:
        identity_id = context.token_data_identity_id or context.prior_identity_id
        for token_data in token_data_by_resource_server.values():
            self._validate_token_data_meets_scope_requirements(
                resource_server=token_data.resource_server,
                token_data=token_data,
                identity_id=identity_id,
            )

    def _validate_token_data_meets_scope_requirements(
        self,
        *,
        resource_server: str,
        token_data: TokenStorageData,
        identity_id: str | None,
        eval_dependent: bool = True,
    ) -> None:
        """
        Evaluate whether the scope requirements for a given resource server are met.

        :param resource_server: A resource server to access scope requirements.
        :param token_data: A token data object to validate.
        :param identity_id: The identity ID of the user, from the surrounding context.
        :param eval_dependent: Whether to evaluate dependent scope requirements.

        :raises UnmetScopeRequirements: If token/consent data does not meet the
            attached root or dependent scope requirements for the resource server.
        """
        required_scopes = self.scope_requirements.get(resource_server)
        # Short circuit - No scope requirements are, by definition, met.
        if required_scopes is None:
            return

        # 1. Does the token meet root scope requirements?
        root_scopes = token_data.scope.split()
        if not all(scope.scope_string in root_scopes for scope in required_scopes):
            raise UnmetScopeRequirementsError(
                "Unmet scope requirements",
                scope_requirements={
                    k: list(v) for k, v in self.scope_requirements.items()
                },
            )

        # Short circuit - No dependent scopes; don't validate them.
        if not eval_dependent or not any(
            scope.dependencies for scope in required_scopes
        ):
            return

        # 2. Does the consent forest meet all dependent scope requirements?
        # 2a. Try with the cached consent forest first.
        forest = self._cached_consent_forest
        if forest is not None and forest.meets_scope_requirements(required_scopes):
            return

        # Identity id is required to fetch consents.
        # if we cannot fetch consents, we cannot do any further validation
        if identity_id is None:
            return

        # 2b. Poll for fresh consents and try again.
        forest = self._poll_and_cache_consents(identity_id)
        if not forest.meets_scope_requirements(required_scopes):
            raise UnmetScopeRequirementsError(
                "Unmet dependent scope requirements",
                scope_requirements={
                    k: list(v) for k, v in self.scope_requirements.items()
                },
            )

    def _poll_and_cache_consents(self, identity_id: str) -> ConsentForest:
        forest = self.consent_client.get_consents(identity_id).to_forest()
        self._cached_consent_forest = forest
        return forest



from __future__ import annotations

import requests

from .retry import (
    RetryCheck,
    RetryCheckFlags,
    RetryCheckResult,
    RetryContext,
    set_retry_check_flags,
)


def check_request_exception(ctx: RetryContext) -> RetryCheckResult:
    """
    Check if a network error was encountered

    :param ctx: The context object which describes the state of the request and the
        retries which may already have been attempted.
    """
    if ctx.exception and isinstance(ctx.exception, requests.RequestException):
        return RetryCheckResult.do_retry
    return RetryCheckResult.no_decision


def check_retry_after_header(ctx: RetryContext) -> RetryCheckResult:
    """
    Check for a retry-after header if the response had a matching status

    :param ctx: The context object which describes the state of the request and the
        retries which may already have been attempted.
    """
    retry_config = ctx.caller_info.retry_config
    if ctx.response is None or (
        ctx.response.status_code not in retry_config.retry_after_status_codes
    ):
        return RetryCheckResult.no_decision
    retry_after = _parse_retry_after(ctx.response)
    if retry_after:
        ctx.backoff = float(retry_after)
    return RetryCheckResult.do_retry


def check_transient_error(ctx: RetryContext) -> RetryCheckResult:
    """
    Check for transient error status codes which could be resolved by retrying
    the request

    :param ctx: The context object which describes the state of the request and the
        retries which may already have been attempted.
    """
    retry_config = ctx.caller_info.retry_config
    if ctx.response is not None and (
        ctx.response.status_code in retry_config.transient_error_status_codes
    ):
        return RetryCheckResult.do_retry
    return RetryCheckResult.no_decision


@set_retry_check_flags(RetryCheckFlags.RUN_ONCE)
def check_expired_authorization(ctx: RetryContext) -> RetryCheckResult:
    """
    This check evaluates whether or not there is invalid or expired authorization
    information which could be updated with some action -- most typically a token
    refresh for an expired access token.

    The check is flagged to only run once per request.

    :param ctx: The context object which describes the state of the request and the
        retries which may already have been attempted.
    """
    retry_config = ctx.caller_info.retry_config
    if (  # is the current check applicable?
        ctx.response is None
        or ctx.caller_info is None
        or ctx.caller_info.authorizer is None
        or (
            ctx.response.status_code
            not in retry_config.expired_authorization_status_codes
        )
    ):
        return RetryCheckResult.no_decision

    # run the authorizer's handler, and 'do_retry' if the handler indicated
    # that it was able to make a change which should make the request retryable
    if ctx.caller_info.authorizer.handle_missing_authorization():
        return RetryCheckResult.do_retry
    return RetryCheckResult.no_decision


def _parse_retry_after(response: requests.Response) -> int | None:
    """
    Get the 'Retry-After' header as an int.

    :param response: The response to parse.
    """
    val = response.headers.get("Retry-After")
    if not val:
        return None
    try:
        return int(val)
    except ValueError:
        return None


DEFAULT_RETRY_CHECKS: tuple[RetryCheck, ...] = (
    check_expired_authorization,
    check_request_exception,
    check_retry_after_header,
    check_transient_error,
)



import globus_sdk
from globus_sdk.testing import load_response


def test_get_task(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_task).metadata
    res = compute_client_v2.get_task(task_id=meta["task_id"])
    assert res.http_status == 200
    assert res.data["task_id"] == meta["task_id"]



import uuid

USER_ID = str(uuid.uuid1())
NON_USER_ID = str(uuid.uuid1())

SUBSCRIPTION_ID = str(uuid.uuid1())

ENDPOINT_ID = str(uuid.uuid1())
ENDPOINT_ID_2 = str(uuid.uuid1())
ENDPOINT_ID_3 = str(uuid.uuid1())

FUNCTION_ID = str(uuid.uuid1())
FUNCTION_ID_2 = str(uuid.uuid1())
FUNCTION_NAME = "howdy_world"
FUNCTION_CODE = "410\n10\n04\n:gASVQAAAAAAAAACMC2hvd2R5X3dvc ..."

TASK_GROUP_ID = str(uuid.uuid1())
TASK_ID = str(uuid.uuid1())
TASK_ID_2 = str(uuid.uuid1())
TASK_DOC = {
    "task_id": TASK_ID,
    "status": "success",
    "result": "10000",
    "completion_t": "1677183605.212898",
    "details": {
        "os": "Linux-5.19.0-1025-aws-x86_64-with-glibc2.35",
        "python_version": "3.10.4",
        "dill_version": "0.3.5.1",
        "globus_compute_sdk_version": "2.3.2",
        "task_transitions": {
            "execution-start": 1692742841.843334,
            "execution-end": 1692742846.123456,
        },
    },
}



import pytest

from globus_sdk import MISSING, GCSAPIError
from globus_sdk.testing import get_last_request, load_response


def test_get_collection_list(client):
    meta = load_response(client.get_collection_list).metadata
    res = client.get_collection_list()

    # check the results against response metadata to ensure integrity and that
    # response parsing works
    assert len(list(res)) == len(meta["collection_ids"])
    assert res["DATA_TYPE"] == "result#1.0.0"
    for index, item in enumerate(res):
        assert item["DATA_TYPE"] == "collection#1.0.0"
        assert item["id"] == meta["collection_ids"][index]
        assert item["storage_gateway_id"] == meta["gateway_ids"][index]
        assert item["display_name"] == meta["display_names"][index]


@pytest.mark.parametrize(
    "include_param, expected",
    (
        (MISSING, None),
        ("foo", "foo"),
        ("foo,bar", "foo,bar"),
        (("foo", "bar"), "foo,bar"),
    ),
)
def test_get_collection_list_include_param(client, include_param, expected):
    load_response(client.get_collection_list)
    client.get_collection_list(include=include_param)
    req = get_last_request()
    if include_param is not MISSING:
        assert "include" in req.params
        assert req.params["include"] == expected
    else:
        assert "include" not in req.params


def test_get_collection_list_mapped_collection_id_param(client):
    load_response(client.get_collection_list)
    client.get_collection_list(mapped_collection_id="MAPPED_COLLECTION")
    assert get_last_request().params.get("mapped_collection_id") == "MAPPED_COLLECTION"


@pytest.mark.parametrize(
    "filter_param, expected",
    (
        (["mapped_collections", "created_by_me"], "mapped_collections,created_by_me"),
        ("created_by_me", "created_by_me"),
    ),
)
def test_get_collection_list_filter_param(client, filter_param, expected):
    load_response(client.get_collection_list)
    client.get_collection_list(filter=filter_param)
    assert get_last_request().params.get("filter") == expected


def test_error_parsing_forbidden(client):
    """
    This test is more focused on error parsing than it is on the actual collection list
    call.
    """
    load_response(client.get_collection_list, case="forbidden")
    with pytest.raises(GCSAPIError) as excinfo:
        client.get_collection_list()

    err = excinfo.value
    assert err.detail is None
    assert err.detail_data_type is None
    assert err.message.startswith("Could not list collections")
    assert err.code == "permission_denied"



from __future__ import annotations

import abc
import typing as t

from .representation import Scope


class ScopeCollection(abc.ABC):
    """
    The common base for scope collections.

    ScopeCollections act as namespaces with attribute access to get scopes.

    They can also be iterated to get all of their defined scopes and provide
    the appropriate resource_server string for use in OAuth2 flows.
    """

    @property
    @abc.abstractmethod
    def resource_server(self) -> str: ...

    @abc.abstractmethod
    def __iter__(self) -> t.Iterator[Scope]: ...


class StaticScopeCollection(ScopeCollection):
    """
    A static scope collection is a data container which provides various scopes
    as class attributes.

    ``resource_server`` must be available as a class attribute.
    """

    resource_server: t.ClassVar[str]

    def __iter__(self) -> t.Iterator[Scope]:
        for view in (vars(self).values(), vars(self.__class__).values()):
            for value in view:
                if isinstance(value, Scope):
                    yield value


class DynamicScopeCollection(ScopeCollection):
    """
    The base type for dynamic scope collections, where the resource server is
    variable.

    The default implementation takes the resource server as the only init-time
    parameter.

    :param resource_server: The resource_server to use for all scopes attached
        to this scope collection.
    """

    # DynamicScopeCollection classes are expected to provide
    # the scope names which they provide as a classvar
    # these are often properties for dynamic computation
    _scope_names: t.ClassVar[tuple[str, ...]]

    def __init__(self, resource_server: str) -> None:
        self._resource_server = resource_server

    def __iter__(self) -> t.Iterator[Scope]:
        for name in self._scope_names:
            value = getattr(self, name)
            if isinstance(value, Scope):
                yield value

    @property
    def resource_server(self) -> str:
        return self._resource_server


def _urn_scope(resource_server: str, scope_name: str) -> Scope:
    """
    Convert a short name + resource server string to a scope, in the Globus
    Auth URN format.

    Example Usage:

    >>> _urn_scope("transfer.api.globus.org", "all")
    Scope('urn:globus:auth:scope:transfer.api.globus.org:all')

    :param resource_server: The resource server string.
    :param scope_name: The short name for the scope.
    """
    return Scope(f"urn:globus:auth:scope:{resource_server}:{scope_name}")


def _url_scope(resource_server: str, scope_name: str) -> Scope:
    """
    Convert a short name + resource server string to a scope, in the Globus
    Auth URL format.

    Example Usage:

    >>> _url_scope("actions.globus.org", "hello_world")
    Scope('https://auth.globus.org/scopes/actions.globus.org/hello_world')

    :param resource_server: The resource server string.
    :param scope_name: The short name for the scope.
    """
    return Scope(f"https://auth.globus.org/scopes/{resource_server}/{scope_name}")



from __future__ import annotations

import typing as t
import uuid

GROUP_ID = str(uuid.uuid1())
USER_ID = str(uuid.uuid1())
SUBSCRIPTION_ID = str(uuid.uuid1())
SUBSCRIPTION_GROUP_ID = str(uuid.uuid1())
SUBSCRIPTION_INFO = {
    "name": "Golems R Us",
    "subscriber_name": "University of Chicago Golem Lab",
    "is_high_assurance": True,
    "is_baa": False,
    "connectors": {
        "052be037-7dda-4d20-b163-3077314dc3e6": {"is_ha": True, "is_baa": False},
        "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72": {"is_ha": True, "is_baa": False},
        "28ef55da-1f97-11eb-bdfd-12704e0d6a4d": {"is_ha": True, "is_baa": False},
        "49b00fd6-63f1-48ae-b27f-d8af4589f876": {"is_ha": True, "is_baa": False},
        "56366b96-ac98-11e9-abac-9cb6d0d9fd63": {"is_ha": True, "is_baa": False},
        "7251f6c8-93c9-11eb-95ba-12704e0d6a4d": {"is_ha": True, "is_baa": False},
        "7643e831-5f6c-4b47-a07f-8ee90f401d23": {"is_ha": True, "is_baa": False},
        "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63": {"is_ha": True, "is_baa": False},
        "7e3f3f5e-350c-4717-891a-2f451c24b0d4": {"is_ha": True, "is_baa": False},
        "9436da0c-a444-11eb-af93-12704e0d6a4d": {"is_ha": True, "is_baa": False},
        "976cf0cf-78c3-4aab-82d2-7c16adbcc281": {"is_ha": True, "is_baa": False},
        "e47b6920-ff57-11ea-8aaa-000c297ab3c2": {"is_ha": True, "is_baa": False},
        "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5": {"is_ha": True, "is_baa": False},
    },
}

BASE_GROUP_DOC: dict[str, t.Any] = {
    "name": "Claptrap's Rough Riders",
    "description": "No stairs allowed.",
    "parent_id": None,
    "id": GROUP_ID,
    "group_type": "regular",
    "enforce_session": False,
    "session_limit": 28800,
    "session_timeouts": {},
    "my_memberships": [
        {
            "group_id": GROUP_ID,
            "identity_id": USER_ID,
            "username": "claptrap@globus.org",
            "role": "member",
            "status": "active",
        }
    ],
    "policies": {
        "group_visibility": "private",
        "group_members_visibility": "managers",
    },
    "subscription_id": None,
    "subscription_info": None,
}

SUBSCRIPTION_GROUP_DOC: dict[str, t.Any] = {
    "name": "University of Chicago, Department of Magical Automata",
    "description": "The finest in machines that go 'ping!' and 'whomp!' and 'bzzt!'",
    "parent_id": None,
    "id": SUBSCRIPTION_GROUP_ID,
    "group_type": "plus",
    "enforce_session": True,
    "session_limit": 1800,
    "session_timeouts": {},
    "my_memberships": [
        {
            "group_id": SUBSCRIPTION_GROUP_ID,
            "identity_id": USER_ID,
            "username": "claptrap@globus.org",
            "role": "member",
            "status": "active",
        }
    ],
    "policies": {
        "group_visibility": "private",
        "group_members_visibility": "managers",
    },
    "subscription_id": SUBSCRIPTION_ID,
    "subscription_info": SUBSCRIPTION_INFO,
}



"""
This module models a read-write object representation of the
X-Globus-Client-Info header.

The spec for X-Globus-Client-Info is documented, in brief, in the GlobusClientInfo class
docstring.
"""

from __future__ import annotations

import typing as t

from globus_sdk import __version__, exc

_RESERVED_CHARS = ";,="


class GlobusClientInfo:
    """
    An implementation of X-Globus-Client-Info as an object. This header encodes a
    mapping of multiple products to versions and potentially other information.

    Values can be added to a clientinfo object via the ``add()`` method.

    The object always initializes itself to start with

        product=python-sdk,version=...

    using the current package version information.

    .. rubric:: ``X-Globus-Client-Info`` Specification

    Header Name: ``X-Globus-Client-Info``

    Header Value:

    - A semicolon (``;``) separated list of client information.

    - Client information is a comma-separated list of ``=`` delimited key-value pairs.
      Well-known values for client-information are:

      - ``product``: A unique identifier of the product.
      - ``version``: Relevant version information for the product.

    - Based on the above, the characters ``;,=`` should be considered reserved and
      should NOT be included in client information values to ensure proper parsing.

    .. rubric:: Example Headers

    .. code-block:: none

        X-Globus-Client-Info: product=python-sdk,version=3.32.1
        X-Globus-Client-Info: product=python-sdk,version=3.32.1;product=cli,version=4.0.0a1

    .. note::

        The ``GlobusClientInfo`` object is not guaranteed to reject all invalid usages.
        For example, ``product`` is required to be unique per header, and users are
        expected to enforce this in their usage.

    :param update_callback: A callback function to be invoked each time the content of
        the GlobusClientInfo changes via ``add()`` or ``clear()``.
    """  # noqa: E501

    def __init__(
        self, *, update_callback: t.Callable[[GlobusClientInfo], None] | None = None
    ) -> None:
        self.infos: list[str] = []

        # set `update_callback` to `None` at first so that `add()` can run without
        # triggering it during `__init__`
        self.update_callback = None
        self.add({"product": "python-sdk", "version": __version__})
        self.update_callback = update_callback

    def __bool__(self) -> bool:
        """Check if there are any values present."""
        return bool(self.infos)

    def format(self) -> str:
        """Format as a header value."""
        return ";".join(self.infos)

    def add(self, value: str | dict[str, str]) -> None:
        """
        Add an item to the clientinfo. The item is either already formatted
        as a string, or is a dict containing values to format.

        :param value: The element to add to the client-info. If it is a dict,
            it may not contain reserved characters in any keys or values. If it is a
            string, it cannot contain the ``;`` separator.
        """
        if not isinstance(value, str):
            value = ",".join(_format_items(value))
        elif ";" in value:
            raise exc.GlobusSDKUsageError(
                "GlobusClientInfo.add() cannot be used to add multiple items in "
                "an already-joined string. Add items separately instead. "
                f"Bad usage: '{value}'"
            )
        self.infos.append(value)

        if self.update_callback is not None:
            self.update_callback(self)

    def clear(self) -> None:
        """Empty the list of info strings and trigger the update callback."""
        self.infos = []
        if self.update_callback is not None:
            self.update_callback(self)


def _format_items(info: dict[str, str]) -> t.Iterable[str]:
    """Format the items in a dict, yielding the contents as an iterable."""
    for key, value in info.items():
        _check_reserved_chars(key, value)
        yield f"{key}={value}"


def _check_reserved_chars(key: str, value: str) -> None:
    """Check a key-value pair to see if it uses reserved chars."""
    if any(c in x for c in _RESERVED_CHARS for x in (key, value)):
        raise exc.GlobusSDKUsageError(
            "X-Globus-Client-Info reserved characters cannot be used in keys or "
            f"values. Bad usage: '{key}: {value}'"
        )



import os

import pytest

from globus_sdk import GlobusSDKUsageError, LocalGlobusConnectPersonal


@pytest.fixture
def pretend_is_windows(monkeypatch):
    """Patch LocalGlobusConnectPersonal to think it's on Windows
    *even if it isn't*.
    """
    monkeypatch.setattr(
        "globus_sdk.local_endpoint.personal.endpoint._on_windows", lambda: True
    )


@pytest.fixture
def pretend_is_not_windows(monkeypatch):
    """Patch LocalGlobusConnectPersonal to think it's NOT on Windows
    *even if it is*.
    """
    monkeypatch.setattr(
        "globus_sdk.local_endpoint.personal.endpoint._on_windows", lambda: False
    )


def test_windows_config_dir_requires_localappdata(pretend_is_windows, monkeypatch):
    monkeypatch.delenv("LOCALAPPDATA", raising=False)
    gcp = LocalGlobusConnectPersonal()
    with pytest.raises(GlobusSDKUsageError):
        gcp.config_dir


def test_windows_config_dir_ignores_localappdata_with_explicit_init(
    pretend_is_windows, monkeypatch
):
    monkeypatch.delenv("LOCALAPPDATA", raising=False)
    gcp = LocalGlobusConnectPersonal(config_dir="/foo/bar")
    assert gcp.config_dir == "/foo/bar"


def test_windows_default_config_dir(pretend_is_windows, monkeypatch, tmp_path):
    monkeypatch.setenv("LOCALAPPDATA", str(tmp_path))
    gcp = LocalGlobusConnectPersonal()
    assert gcp.config_dir == str(tmp_path / "Globus Connect")


def test_nonwindows_default_config_dir(pretend_is_not_windows, monkeypatch, tmp_path):
    def _fake_expanduser(p: str):
        assert p.startswith("~/")
        return str(tmp_path / p[2:])

    monkeypatch.setattr(os.path, "expanduser", _fake_expanduser)
    gcp = LocalGlobusConnectPersonal()
    assert gcp.config_dir == str(tmp_path / ".globusonline")



from globus_sdk.testing import get_last_request, load_response


def test_delete_tunnel(client):
    meta = load_response(client.delete_tunnel).metadata

    res = client.delete_tunnel(meta["tunnel_id"])
    assert res.http_status == 200

    req = get_last_request()
    assert req.body is None



import typing as t

import pytest

import globus_sdk


@pytest.fixture
def flows_client():
    client = globus_sdk.FlowsClient()
    with client.retry_config.tune(max_retries=0):
        yield client


@pytest.fixture
def specific_flow_client_class() -> t.Type[globus_sdk.SpecificFlowClient]:
    class CustomSpecificFlowClient(globus_sdk.SpecificFlowClient):
        def __init__(self, **kwargs) -> None:
            super().__init__(**kwargs)
            self.retry_config.max_retries = 0

    return CustomSpecificFlowClient



import uuid

import pytest

from globus_sdk.testing import get_last_request
from tests.common import register_api_route

ZERO_ID = uuid.UUID(int=0)


def get_last_params():
    return get_last_request().params


@pytest.fixture
def task_id():
    return uuid.uuid1()


# stub in empty data, this can be explicitly replaced if a test wants specific data
@pytest.fixture(autouse=True)
def empty_response(task_id):
    register_api_route(
        "transfer", f"/endpoint_manager/task/{task_id}/event_list", json={"DATA": []}
    )


# although int values are not supported based on the type annotations, users may already
# be passing ints -- it's good to support this usage and test it even if it's not
# documented and desirable
@pytest.mark.parametrize(
    "paramvalue, paramstr", [(True, "1"), (False, "0"), (1, "1"), (0, "0")]
)
def test_filter_is_error(client, task_id, paramvalue, paramstr):
    client.endpoint_manager_task_event_list(task_id, filter_is_error=paramvalue)
    params = get_last_params()
    assert "filter_is_error" in params
    assert params["filter_is_error"] == paramstr



from .client import TimersClient
from .data import (
    FlowTimer,
    OnceTimerSchedule,
    RecurringTimerSchedule,
    TimerJob,
    TransferTimer,
)
from .errors import TimersAPIError

__all__ = (
    "FlowTimer",
    "TimersAPIError",
    "TimersClient",
    "OnceTimerSchedule",
    "RecurringTimerSchedule",
    "TimerJob",
    "TransferTimer",
)



"""
This is a very small 'tox' plugin.
'toxfile.py' is a special name for auto-loading a plugin without defining package
metadata.

For full doc, see: https://tox.wiki/en/latest/plugins.html

Methods decorated below with `tox.plugin.impl` are hook implementations.
We only implement hooks which we need.
"""

from __future__ import annotations

import logging
import pathlib
import shutil
import typing as t

from tox.plugin import impl

if t.TYPE_CHECKING:
    from tox.config.sets import EnvConfigSet
    from tox.session.state import State
    from tox.tox_env.api import ToxEnv

log = logging.getLogger(__name__)

REPO_ROOT = pathlib.Path(__file__).parent
BUILD_DIR = REPO_ROOT / "build"


@impl
def tox_on_install(
    tox_env: ToxEnv, arguments: t.Any, section: str, of_type: str
) -> None:
    # when setting up the wheel build, clear the 'build/' directory if present
    if tox_env.name != "build_wheel":
        return
    if BUILD_DIR.exists():
        shutil.rmtree(BUILD_DIR)


@impl
def tox_add_env_config(env_conf: EnvConfigSet, state: State) -> None:
    env_conf.add_config(
        keys=["globus_sdk_rmtree"],
        of_type=list[str],
        default=[],
        desc="A dir tree to remove before running the environment commands",
    )


@impl
def tox_before_run_commands(tox_env: ToxEnv) -> None:
    # determine if it was a parallel invocation by examining the CLI command
    parallel_detected = tox_env.options.command in ("p", "run-parallel")
    if parallel_detected:
        # tox is running parallel, set an indicator env var
        # and effectively disable pytest-xdist by setting xdist-workers to 0
        # -- 0 means tests will run in the main process, not even in a worker
        setenv = tox_env.conf.load("set_env")
        setenv.update({"TOX_PARALLEL": "1", "PYTEST_XDIST_AUTO_NUM_WORKERS": "0"})

    sdk_rmtree = tox_env.conf.load("globus_sdk_rmtree")
    for name in sdk_rmtree:
        path = pathlib.Path(name)
        if path.exists():
            log.warning(f"globus_sdk_rmtree: {path}")
            shutil.rmtree(path)



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

APP_REQUEST_ARGS = {
    "template_id": str(uuid.uuid1()),
    "name": str(uuid.uuid1()).replace("-", ""),
}


def make_app_request_body(request_args: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    request_body = request_args.copy()
    request_body["template_id"] = str(request_args["template_id"])
    return request_body


def make_app_response_body(request_args: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    return {
        "client": {
            "fqdns": [],
            "name": request_args["name"],
            "id": "e634cc2a-d528-494e-8dda-92ec54a883c9",
            "public_client": False,
            "scopes": [],
            "required_idp": None,
            "grant_types": [
                "authorization_code",
                "client_credentials",
                "refresh_token",
            ],
            "userinfo_from_effective_identity": True,
            "client_type": "confidential_client",
            "prompt_for_named_grant": False,
            "links": {"privacy_policy": None, "terms_and_conditions": None},
            "visibility": "private",
            "preselect_idp": None,
            "parent_client": str(request_args["template_id"]),
            "project": None,
            "redirect_uris": [],
        },
        "included": {
            "client_credential": {
                "name": "Auto-created at client creation",
                "id": "b4840855-2de8-4035-b1b4-4e7c8f518943",
                "client": "e634cc2a-d528-494e-8dda-92ec54a883c9",
                "secret": "cgK1HG9Y0DcZw79YlQEJpZCF4CMxIbaFf5sohWxjcfY=",
            }
        },
    }


def register_response(
    args: t.Mapping[str, t.Any],
) -> RegisteredResponse:
    request_args = {**APP_REQUEST_ARGS, **args}
    request_body = make_app_request_body(request_args)
    response_body = make_app_response_body(request_args)

    return RegisteredResponse(
        service="auth",
        method="POST",
        path="/v2/api/clients",
        json={"client": response_body},
        metadata={
            # Test functions use 'args' to form request
            "args": request_args,
            # Test functions use 'response' to verify response
            "response": response_body,
        },
        match=[
            json_params_matcher(
                {"client": request_body},
            )
        ],
    )


RESPONSES = ResponseSet(
    default=register_response({}),
    template_id_str=register_response({"template_id": str(uuid.uuid1())}),
    template_id_uuid=register_response({"template_id": uuid.uuid1()}),
    name=register_response({"name": str(uuid.uuid1()).replace("-", "")}),
)



from .app import GlobusApp
from .client_app import ClientApp
from .config import GlobusAppConfig
from .protocols import (
    IDTokenDecoderProvider,
    LoginFlowManagerProvider,
    TokenStorageProvider,
    TokenValidationErrorHandler,
)
from .user_app import UserApp

__all__ = (
    "GlobusApp",
    "UserApp",
    "ClientApp",
    "GlobusAppConfig",
    # Protocols
    "IDTokenDecoderProvider",
    "TokenValidationErrorHandler",
    "TokenStorageProvider",
    "LoginFlowManagerProvider",
)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import UNAUTHORIZED_AUTH_RESPONSE

_globus_at_globus_data = {
    "email": None,
    "id": "46bd0f56-e24f-11e5-a510-131bef46955c",
    "identity_provider": "7daddf46-70c5-45ee-9f0f-7244fe7c8707",
    "name": None,
    "organization": None,
    "status": "unused",
    "username": "globus@globus.org",
}
_sirosen_at_globus_data = {
    "email": "sirosen@globus.org",
    "id": "ae341a98-d274-11e5-b888-dbae3a8ba545",
    "identity_provider": "927d7238-f917-4eb2-9ace-c523fa9ba34e",
    "name": "Stephen Rosen",
    "organization": "Globus Team",
    "status": "used",
    "username": "sirosen@globus.org",
}
_globus_at_globusid_data = {
    "email": "support@globus.org",
    "id": str(uuid.UUID(int=1)),
    "identity_provider": "41143743-f3c8-4d60-bbdb-eeecaba85bd9",
    "identity_type": "login",
    "name": "Globus Team",
    "organization": "University of Chicago",
    "status": "used",
    "username": "globus@globusid.org",
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/api/identities",
        json={"identities": [_globus_at_globus_data]},
        metadata={
            "id": _globus_at_globus_data["id"],
            "username": _globus_at_globus_data["username"],
        },
    ),
    empty=RegisteredResponse(
        service="auth",
        path="/v2/api/identities",
        json={"identities": []},
    ),
    multiple=RegisteredResponse(
        service="auth",
        path="/v2/api/identities",
        json={"identities": [_globus_at_globus_data, _sirosen_at_globus_data]},
        metadata={
            "ids": [_globus_at_globus_data["id"], _sirosen_at_globus_data["id"]],
            "usernames": [
                _globus_at_globus_data["username"],
                _sirosen_at_globus_data["username"],
            ],
        },
    ),
    globusid=RegisteredResponse(
        service="auth",
        path="/v2/api/identities",
        json={"identities": [_globus_at_globusid_data]},
        metadata={
            "id": _globus_at_globusid_data["id"],
            "username": _globus_at_globusid_data["username"],
            "short_username": _globus_at_globusid_data["username"].partition("@")[0],
            "org": _globus_at_globusid_data["organization"],
        },
    ),
    sirosen=RegisteredResponse(
        service="auth",
        path="/v2/api/identities",
        json={"identities": [_sirosen_at_globus_data]},
        metadata={
            "id": _sirosen_at_globus_data["id"],
            "username": _sirosen_at_globus_data["username"],
            "org": _sirosen_at_globus_data["organization"],
        },
    ),
    unauthorized=RegisteredResponse(
        service="auth",
        path="/v2/api/identities",
        status=401,
        json=UNAUTHORIZED_AUTH_RESPONSE.json,
        metadata={"error_id": UNAUTHORIZED_AUTH_RESPONSE.error_id},
    ),
)



from globus_sdk.testing import load_response


def test_get_clients(service_client):
    meta = load_response(service_client.get_clients).metadata
    res = service_client.get_clients()

    assert {client["id"] for client in res["clients"]} == set(meta["client_ids"])



from __future__ import annotations

import uuid

import pytest

from globus_sdk import GlobusSDKUsageError
from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "case_name",
    (
        "name",
        "publicly_visible",
        "not_publicly_visible",
        "redirect_uris",
        "links",
        "required_idp",
        "preselect_idp",
    ),
)
def test_update_client_args(
    service_client,
    case_name: str,
):
    meta = load_response(service_client.update_client, case=case_name).metadata

    res = service_client.update_client(**meta["args"])
    for k, v in meta["response"].items():
        assert res["client"][k] == v


def test_links_requirement(service_client):
    """
    Verify that terms_and_conditions and privacy_policy must be used together.
    """
    with pytest.raises(GlobusSDKUsageError):
        service_client.create_client(
            "FOO",
            uuid.uuid1(),
            visibility="public",
            terms_and_conditions="https://foo.net",
        )

    with pytest.raises(GlobusSDKUsageError):
        service_client.create_client(
            "FOO",
            uuid.uuid1(),
            visibility="public",
            privacy_policy="https://foo.net",
        )



from __future__ import annotations

import datetime
import sys
import typing as t

import jwt
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from globus_sdk.response import GlobusHTTPResponse

from ._common import SupportsJWKMethods

if t.TYPE_CHECKING:
    from globus_sdk import AuthLoginClient, GlobusAppConfig

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


class IDTokenDecoder:
    """
    JWT decoder for OIDC ID tokens issued by Globus Auth.
    Decoding uses a client object to fetch necessary data from Globus Auth.

    By default, the OIDC configuration data and JWKs will be cached in an internal dict.
    An alternative cache can be provided on init to use an alternative storage
    mechanism.

    The ``get_jwt_audience`` and ``get_jwt_leeway`` methods supply parameters to
    decoding. Subclasses can override these methods to customize the decoder.

    :param auth_client: The client which should be used to callout to Globus Auth as
        needed. Any AuthClient or AuthLoginClient will work for this purpose.
    :param jwt_leeway: The JWT leeway to use during decoding, as a number of seconds
        or a timedelta. The default is 5 minutes.
    :param jwt_options: The ``options`` passed to the underlying JWT decode function.
        Defaults to an empty dict.
    """

    def __init__(
        self,
        auth_client: SupportsJWKMethods,
        *,
        # default to 300 seconds
        #
        # valuable inputs to this number:
        # - expected clock drift per day (6s for a bad clock)
        # - Windows time sync interval (64s)
        # - Windows' stated goal of meeting the Kerberos 5 clock skew requirement (5m)
        # - ntp panic threshold (1000s of drift)
        # - the knowledge that VM clocks typically run slower and may skew significantly
        #
        # NTP panic should be understood as a critical error; 1000s of drift is
        # therefore too high for us to allow.
        #
        # 300s (5m) is therefore chosen to match the Windows desired maximum for
        # clock drift, and the underlying Kerberos requirement.
        jwt_leeway: float | datetime.timedelta = 300.0,
        jwt_options: dict[str, t.Any] | None = None,
    ) -> None:
        self._auth_client = auth_client
        self._openid_configuration: dict[str, t.Any] | None = None
        self._jwk: RSAPublicKey | None = None

        self.jwt_leeway: float | datetime.timedelta = jwt_leeway
        self.jwt_options: dict[str, t.Any] = (
            jwt_options if jwt_options is not None else {}
        )

    @classmethod
    def for_globus_app(
        cls,
        *,
        app_name: str,  # pylint: disable=unused-argument
        config: GlobusAppConfig,  # pylint: disable=unused-argument
        login_client: AuthLoginClient,
    ) -> Self:
        """
        Create an ``IDTokenDecoder`` for use in a GlobusApp.

        :param app_name: The name supplied to the GlobusApp.
        :param config: The configuration supplied to the GlobusApp.
        :param login_client: A login client to use for instantiating an
            ``IDTokenDecoder``.
        """
        return cls(login_client)

    def decode(self, id_token: str, /) -> dict[str, t.Any]:
        """
        The ``decode()`` method takes an ``id_token`` as a string, and decodes it to
        a dictionary.

        This method should implicitly retrieve the OpenID configuration and JWK
        for Globus Auth.

        :param id_token: The token to decode
        """
        audience = self.get_jwt_audience()
        openid_configuration = self.get_openid_configuration()
        jwk = self.get_jwk()

        signing_algos = openid_configuration["id_token_signing_alg_values_supported"]

        return jwt.decode(
            id_token,
            key=jwk,
            algorithms=signing_algos,
            audience=audience,
            options=self.jwt_options,
            leeway=self.jwt_leeway,
        )

    def get_jwt_audience(self) -> str | None:
        """
        The audience for JWT verification defaults to the client's client ID.
        """
        return self._auth_client.client_id

    def store_openid_configuration(
        self, openid_configuration: dict[str, t.Any] | GlobusHTTPResponse
    ) -> None:
        """
        Store openid_configuration data for future use in ``decode()``.

        :param openid_configuration: The configuration data, as fetched via
            :meth:`AuthClient.get_openid_configuration`
        """
        if isinstance(openid_configuration, GlobusHTTPResponse):
            self._openid_configuration = openid_configuration.data
        else:
            self._openid_configuration = openid_configuration

    def store_jwk(self, jwk: RSAPublicKey) -> None:
        """
        Store a JWK for future use in ``decode()``.

        :param jwk: The JWK, as fetched via :meth:`AuthClient.get_jwk`
            with ``as_pem=True``.
        """
        self._jwk = jwk

    def get_openid_configuration(self) -> dict[str, t.Any]:
        """
        Fetch the OpenID Configuration for Globus Auth, and cache the result before
        returning it.

        If a config was previously stored, return that instead.
        """
        if self._openid_configuration is None:
            self._openid_configuration = (
                self._auth_client.get_openid_configuration().data
            )
        return self._openid_configuration

    def get_jwk(self) -> RSAPublicKey:
        """
        Fetch the JWK for Globus Auth, and cache the result before returning it.

        If a key was previously stored, return that instead.
        """
        if not self._jwk:
            self._jwk = self._auth_client.get_jwk(
                openid_configuration=self.get_openid_configuration(), as_pem=True
            )
        return self._jwk



import pytest

from globus_sdk.token_storage.legacy import SQLiteAdapter


def test_sqlite_reading_bad_config():
    adapter = SQLiteAdapter(":memory:")
    # inject bad data (array, needs to be dict)
    # store_config does not check the input type, just uses json.dumps()
    adapter.store_config("foo_conf", [])

    with pytest.raises(ValueError, match="reading config data and got non-dict result"):
        adapter.read_config("foo_conf")
    adapter.close()


def test_sqlite_reading_bad_token_data():
    adapter = SQLiteAdapter(":memory:")
    # inject bad data (array, needs to be dict)
    adapter._connection.execute(
        """\
INSERT INTO token_storage(namespace, resource_server, token_data_json)
VALUES (?, ?, ?)""",
        (adapter.namespace, "foo_rs", "[]"),
    )
    with pytest.raises(
        ValueError, match="data error: token data was not saved as a dict"
    ):
        adapter.get_token_data("foo_rs")
    adapter.close()


def test_sqliteadapter_passes_connect_params():
    with pytest.raises(TypeError):
        SQLiteAdapter(":memory:", connect_params={"invalid_kwarg": True})

    adapter = SQLiteAdapter(":memory:", connect_params={"timeout": 10})
    adapter.close()



import copy
import pickle

import pytest

from globus_sdk._missing import MISSING, MissingType


def test_missing_type_cannot_be_instantiated():
    with pytest.raises(TypeError, match="MissingType should not be instantiated"):
        MissingType()


def test_missing_sentinel_bools_as_false():
    assert bool(MISSING) is False


def test_str_of_missing():
    assert str(MISSING) == "<globus_sdk.MISSING>"


def test_copy_of_missing_is_self():
    assert copy.copy(MISSING) is MISSING
    assert copy.deepcopy(MISSING) is MISSING


def test_pickle_of_missing_is_self():
    assert pickle.loads(pickle.dumps(MISSING)) is MISSING



from globus_sdk.testing import get_last_request, load_response


def test_list_stream_access_points(client):
    meta = load_response(client.list_stream_access_points).metadata

    res = client.list_stream_access_points()
    assert res.http_status == 200
    list_results = list(res)

    assert len(list_results) == 2
    for stream_access_point in list_results:
        assert stream_access_point["type"] == "StreamAccessPoint"
        assert stream_access_point["id"] in meta["access_point_ids"]

    req = get_last_request()
    assert req.body is None



import globus_sdk
from globus_sdk.testing import load_response


def test_get_endpoint(client):
    meta = load_response(client.get_endpoint).metadata

    get_response = client.get_endpoint()

    assert get_response["DATA_TYPE"] == "endpoint#1.2.0"
    assert get_response["display_name"] == meta["display_name"]
    assert get_response["id"] == meta["endpoint_id"]


def test_update_endpoint(client):
    meta = load_response(client.update_endpoint).metadata

    update_doc = globus_sdk.EndpointDocument(display_name=meta["display_name"])
    update_response = client.update_endpoint(update_doc)

    assert update_response["DATA_TYPE"] == "result#1.0.0"
    assert update_response["message"] == f"Updated endpoint {meta['endpoint_id']}"


def test_update_endpoint_including_endpoint_in_response(client):
    meta = load_response(client.update_endpoint, case="include_endpoint").metadata

    update_doc = globus_sdk.EndpointDocument(display_name=meta["display_name"])
    update_response = client.update_endpoint(update_doc, include=["endpoint"])

    assert update_response["DATA_TYPE"] == "endpoint#1.2.0"
    assert update_response["display_name"] == meta["display_name"]
    assert update_response["id"] == meta["endpoint_id"]


def test_endpoint_document_infers_data_type():
    doc = globus_sdk.EndpointDocument(display_name="My Endpoint")

    assert doc["DATA_TYPE"] == "endpoint#1.0.0"


def test_endpoint_document_infers_data_type_when_control_port_specified():
    doc = globus_sdk.EndpointDocument(
        display_name="My Endpoint", gridftp_control_channel_port=2811
    )

    assert doc["DATA_TYPE"] == "endpoint#1.1.0"


def test_endpoint_document_respects_explicit_data_type():
    doc = globus_sdk.EndpointDocument(
        data_type="endpoint#0.15.x", display_name="My Endpoint"
    )

    assert doc["DATA_TYPE"] == "endpoint#0.15.x"



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TIMER_ID, V1_TIMER

RESPONSES = ResponseSet(
    metadata={"job_ids": [TIMER_ID]},
    default=RegisteredResponse(
        service="timer",
        path="/jobs/",
        method="GET",
        json={"jobs": [V1_TIMER]},
    ),
)



import uuid
from collections import namedtuple

_ErrorData = namedtuple("_ErrorData", ("error_id", "json", "metadata_include"))

_error_id = str(uuid.uuid1())
UNAUTHORIZED_AUTH_RESPONSE = _ErrorData(
    _error_id,
    {
        "errors": [
            {
                "status": "401",
                "id": _error_id,
                "code": "UNAUTHORIZED",
                "detail": "Call must be authenticated",
                "title": "Unauthorized",
            }
        ],
        "error_description": "Unauthorized",
        "error": "unauthorized",
    },
    {
        "http_status": 401,
        "code": "UNAUTHORIZED",
        "message": "Call must be authenticated",
    },
)

_error_id = str(uuid.uuid1())
FORBIDDEN_AUTH_RESPONSE = _ErrorData(
    _error_id,
    {
        "errors": [
            {
                "status": "403",
                "id": _error_id,
                "code": "FORBIDDEN",
                "detail": "Call must be authenticated",
                "title": "Unauthorized",
            }
        ],
        "error_description": "Unauthorized",
        "error": "unauthorized",
    },
    {
        "http_status": 403,
        "code": "FORBIDDEN",
        "message": "Call must be authenticated",
    },
)



from __future__ import annotations

import ast
import typing as t

CODEMAP: dict[str, str] = {
    # SDK001 is necessary for SDK002 enforcement to be easy
    # otherwise, we would have to have a more sophisticated linter which knows about
    # lexical scopes!
    "SDK001": "SDK001 loggers should be named 'log'",
    "SDK002": "SDK002 never use 'log.info'",
    # don't do `isinstance(x, MissingType)` -- use `x is MISSING` instead
    "SDK003": "SDK003 use `is MISSING`, not `isinstance(..., MissingType)`",
}


class Plugin:
    name = "globus-sdk-flake8"
    version = "1.0.0"

    # args to init determine plugin behavior. see:
    # https://flake8.pycqa.org/en/latest/plugin-development/plugin-parameters.html#indicating-desired-data
    #
    # by having "tree" as an init arg, we tell flake8 that we are an AST-handling
    # plugin, run once per file
    def __init__(self, tree: ast.AST) -> None:
        self.tree = tree

    # Plugin.run() is how checks will run. For detail, see implementation of:
    # https://flake8.pycqa.org/en/latest/internal/checker.html#flake8.checker.FileChecker.run_ast_checks
    def run(self) -> t.Iterator[tuple[int, int, str, type]]:
        visitor = SDKVisitor()
        visitor.visit(self.tree)
        for lineno, col, code in visitor.collect:
            yield lineno, col, CODEMAP[code], type(self)


class SDKVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.collect: list[tuple[int, int, str]] = []

    def _record(self, node: ast.expr | ast.stmt, code: str) -> None:
        self.collect.append((node.lineno, node.col_offset, code))

    def visit_Assign(self, node: ast.Assign) -> None:
        if matches_sdk001(node):
            self._record(node, "SDK001")

        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        if matches_sdk003(node):
            self._record(node, "SDK003")
        elif matches_sdk002(node):
            self._record(node, "SDK002")

        self.generic_visit(node)


def matches_sdk001(node: ast.Assign) -> bool:
    """
    A matcher for the SDK001 lint rule.

    Checks for `x = logging.getLogger(__name__)` where `x` is not `log`.

    :param node: the assignment statement AST node to check
    """
    # the value must be a call and must be using attr access in the function call
    # this eliminates bare funcs like `x = foo()` but allows `x = foo.bar()`
    if not (
        isinstance(node.value, ast.Call) and isinstance(node.value.func, ast.Attribute)
    ):
        return False

    call_node: ast.Call = node.value
    func_node: ast.Attribute = node.value.func

    # make sure it's 'logging.getLogger' and no other function
    if not (
        func_node.attr == "getLogger"
        and isinstance(func_node.value, ast.Name)
        and func_node.value.id == "logging"
    ):
        return False

    # the assignee must be a single variable and it must be a name node
    if not (len(node.targets) == 1 and isinstance(node.targets[0], ast.Name)):
        return False

    # if the assigned name is `log`, then it cannot be a match
    name_node: ast.Name = node.targets[0]
    if name_node.id == "log":
        return False

    # confirm that the `logging.getLogger` args are a single name -- that's the
    # form it will be when `__name__` is passed
    if len(call_node.args) != 1 or not isinstance(call_node.args[0], ast.Name):
        return False

    # if the argument is some other name, e.g. `logging.getLogger(my_variable)`,
    # then that cannot be a match
    logger_arg: ast.Name = call_node.args[0]
    if logger_arg.id != "__name__":
        return False

    return True  # all conditions met!


def matches_sdk002(node: ast.Call) -> bool:
    """
    A matcher for the SDK002 lint rule.

    Checks for `log.info(...)`

    :param node: the function call AST node to check
    """
    func_node = node.func
    # the function call must be of the form 'OBJ.info(...)'
    if not (isinstance(func_node, ast.Attribute) and func_node.attr == "info"):
        return False

    # and, more specifically, the object must be named 'log', so
    # it's `log.info(...)
    if not (isinstance(func_node.value, ast.Name) and func_node.value.id == "log"):
        return False

    return True  # all conditions met!


def matches_sdk003(node: ast.Call) -> bool:
    """
    A matcher for the SDK003 lint rule.

    Checks for `isinstance(x, MissingType)`

    :param node: the function call AST node to check
    """
    # it must be a call to a function named "isinstance"
    if not (isinstance(node.func, ast.Name) and node.func.id == "isinstance"):
        return False
    # if the number of arguments is improper, it can't be a violation (not a
    # valid 'isinstance' call)
    if len(node.args) != 2:
        return False
    right_hand_side = node.args[1]
    # the second argument must be the name 'MissingType'
    if not (
        isinstance(right_hand_side, ast.Name) and right_hand_side.id == "MissingType"
    ):
        return False

    return True  # all conditions met!



import pytest

import globus_sdk


def test_explicit_dir_func_works():
    assert "TransferClient" in dir(globus_sdk)
    assert "__all__" in dir(globus_sdk)


def test_force_eager_imports_can_run():
    # this check will not do much, other than ensuring that this does not crash
    globus_sdk._force_eager_imports()


def test_attribute_error_on_bad_name():
    with pytest.raises(AttributeError) as excinfo:
        globus_sdk.DEIMOS_DOWN_REMOVE_ALL_PLANTS

    err = excinfo.value
    assert (
        str(err) == "module globus_sdk has no attribute DEIMOS_DOWN_REMOVE_ALL_PLANTS"
    )



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token/revoke",
        method="POST",
        json={"active": False},
    )
)



import pytest

import globus_sdk


@pytest.fixture
def client_class():
    class CustomClient(globus_sdk.BaseClient):
        service_name = "foo"

    return CustomClient


@pytest.fixture
def client(client_class):
    return client_class()



from .command_line_login_flow_manager import (
    CommandLineLoginFlowEOFError,
    CommandLineLoginFlowManager,
)
from .local_server_login_flow_manager import (
    LocalServerEnvironmentalLoginError,
    LocalServerLoginError,
    LocalServerLoginFlowManager,
)
from .login_flow_manager import LoginFlowManager

__all__ = (
    "CommandLineLoginFlowManager",
    "CommandLineLoginFlowEOFError",
    "LocalServerLoginError",
    "LocalServerEnvironmentalLoginError",
    "LocalServerLoginFlowManager",
    "LoginFlowManager",
)
